"""diverse_seq: a tool for sampling diverse biological sequences"""

import os
import typing

# need to import hdf5plugin here to make sure the plugin path can be
# found by h5py
import hdf5plugin  # noqa: F401

os.environ["COGENT3_NEW_TYPE"] = "1"

if typing.TYPE_CHECKING:
    from cogent3.core.new_alignment import SequenceCollection

__version__ = "2025.6.26"


def load_sample_data() -> "SequenceCollection":
    """load sample data"""
    from cogent3 import load_aligned_seqs

    from .util import get_sample_data_path

    path = get_sample_data_path()
    return load_aligned_seqs(path, moltype="dna", new_type=True).degap()



from pathlib import Path

import pytest
from cogent3 import load_unaligned_seqs
from cogent3.core.new_alignment import SequenceCollection


@pytest.fixture(scope="session")
def DATA_DIR() -> Path:
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def seq_path(DATA_DIR: Path) -> Path:
    return DATA_DIR / "brca1.fasta"


@pytest.fixture
def unaligned_seqs(seq_path: Path) -> SequenceCollection:
    return load_unaligned_seqs(seq_path, moltype="dna")


@pytest.fixture
def tmp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("dvgt")



import random
import sys
import tempfile
import time
import typing
from collections import OrderedDict
from collections.abc import Mapping
from pathlib import Path

import click
import numpy
import rich.progress as rich_progress
from cogent3.app import data_store as c3_data_store
from scitrack import CachingLogger

from diverse_seq import __version__
from diverse_seq import cluster as dvs_cluster
from diverse_seq import data_store as dvs_data_store
from diverse_seq import io as dvs_io
from diverse_seq import records as dvs_records
from diverse_seq import util as dvs_util

if typing.TYPE_CHECKING:
    from click.core import Context, Option

LOGGER = CachingLogger()


def _get_seed(ctx: "Context", param: "Option", value: str | None) -> int:
    value = time.time() if value is None else value
    return int(value)


# custom group class to ensure help function returns commands in desired order.
# class is adapted from Максим Стукало's answer to
# https://stackoverflow.com/questions/47972638/how-can-i-define-the-order-of-click-sub-commands-in-help


class OrderedGroup(click.Group):
    def __init__(
        self,
        name: str | None = None,
        commands: Mapping[str, click.Command] | None = None,
        **kwargs,
    ):
        super().__init__(name, commands, **kwargs)
        #: the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.commands


_click_command_opts = {
    "no_args_is_help": True,
    "context_settings": {"show_default": True},
}


@click.group(cls=OrderedGroup)
@click.version_option(__version__)  # add version option
def main() -> None:
    """dvs -- alignment free detection of the most diverse sequences using JSD"""


_hide_progress = click.option(
    "-hp",
    "--hide_progress",
    is_flag=True,
    callback=dvs_util._hide_progress,
    help="hide progress bars",
)
_verbose = click.option(
    "-v",
    "--verbose",
    count=True,
    help="is an integer indicating number of cl occurrences",
)
_outpath = click.option(
    "-o",
    "--outpath",
    type=Path,
    help="path to write output file",
    required=True,
)
_overwrite = click.option(
    "-F",
    "--force_overwrite",
    is_flag=True,
    default=False,
    help="Overwrite existing file if it exists",
)
_include = click.option(
    "-i",
    "--include",
    callback=dvs_util._comma_sep_or_file,  # noqa: SLF001
    help="seqnames to include in divergent set",
)
_suffix = click.option("-sf", "--suffix", default="fa", help="sequence file suffix")
_numprocs = click.option("-np", "--numprocs", default=1, help="number of processes")
_limit = click.option("-L", "--limit", type=int, help="number of sequences to process")
_moltype = click.option(
    "-m",
    "--moltype",
    type=click.Choice(["dna", "rna"]),
    default="dna",
    help="Molecular type of sequences",
)
_seqfile = click.option(
    "-s",
    "--seqfile",
    required=True,
    type=Path,
    callback=dvs_util._check_h5_dstore,
    help="path to .dvseqs file",
)
_k = click.option("-k", type=int, default=6, help="k-mer size")

_seed = click.option(
    "--seed",
    type=int,
    callback=_get_seed,
    help="seed for random number generator, defaults to system clock",
)

_click_opts = _click_command_opts.copy()
_click_opts.pop("no_args_is_help")


@main.command(**_click_opts)
@click.option(
    "-o",
    "--outpath",
    type=Path,
    default=Path("demo.fa"),
    help="write a demo fasta file",
)
def demo_data(outpath: Path) -> None:
    """Export a demo sequence file"""
    from diverse_seq import load_sample_data

    seqs = load_sample_data()
    seqs.write(outpath, file_format="fasta")
    dvs_util.print_colour(f"Wrote '{outpath!s}'", "green")


@main.command(**_click_command_opts)
@click.option(
    "-s",
    "--seqdir",
    required=True,
    type=Path,
    help="one sequence file, or a directory containing multiple sequence files",
)
@_suffix
@click.option(
    "-o",
    "--outpath",
    required=True,
    type=Path,
    help="write processed seqs to this filename",
)
@_numprocs
@_overwrite
@_moltype
@_limit
@_hide_progress
def prep(
    seqdir: Path,
    suffix: str,
    outpath: Path,
    numprocs: int,
    force_overwrite: bool,
    moltype: str,
    limit: int | None,
    hide_progress: bool,
) -> None:
    """Writes processed sequences to a <HDF5 file>.dvseqs."""
    dvseqs_path = outpath.with_suffix(".dvseqs")
    if dvseqs_path.exists() and not force_overwrite:
        dvs_util.print_colour(
            "A file with the same name already exists. Existing data members will be skipped. "
            "Use the -F flag if you want to overwrite the existing file.",
            "blue",
        )
        sys.exit(1)

    if dvseqs_path.exists() and force_overwrite:
        dvseqs_path.unlink()

    suffix = suffix.removeprefix(".")

    seq_format = dvs_util.get_seq_file_format(suffix)
    if seq_format is None:
        dvs_util.print_colour(
            f"Unrecognised sequence file suffix '{suffix}'",
            "red",
        )
        sys.exit(1)

    with dvs_util.keep_running(), tempfile.TemporaryDirectory() as tmp_dir:
        if seqdir.is_file():
            convert2dstore = dvs_io.dvs_file_to_dir(dest=tmp_dir)
            in_dstore = convert2dstore(seqdir)  # pylint: disable=not-callable
        else:
            in_dstore = c3_data_store.DataStoreDirectory(source=seqdir, suffix=suffix)

        if len(in_dstore) < 5:
            msg = f"Num files matching '{seqdir}/*.{suffix}' = {len(in_dstore)} < 5."
            if seqdir.is_dir():
                msg = f"{msg} Did you mean to pass a file path instead?"
            dvs_util.print_colour(msg, "red")
            sys.exit(1)

        if limit is not None:
            members = in_dstore.completed[:]
            random.shuffle(members)
            in_dstore = members[:limit]

        out_dstore = dvs_data_store.HDF5DataStore(source=dvseqs_path, mode="w")

        loader = dvs_io.dvs_load_seqs(
            moltype=moltype,
            seq_format=seq_format,
        )
        writer = dvs_io.dvs_write_seqs(
            data_store=out_dstore,
        )

        with rich_progress.Progress(
            rich_progress.TextColumn("[progress.description]{task.description}"),
            rich_progress.BarColumn(),
            rich_progress.TaskProgressColumn(),
            rich_progress.TimeRemainingColumn(),
            rich_progress.TimeElapsedColumn(),
            disable=hide_progress,
        ) as progress:
            convert = progress.add_task("Processing sequences", total=len(in_dstore))
            for r in loader.as_completed(
                in_dstore,
                show_progress=False,
                parallel=numprocs > 1,
                par_kw={"max_workers": numprocs},
            ):
                if not r:
                    dvs_util.print_colour(str(r), "red")
                    sys.exit(1)
                writer(r)  # pylint: disable=not-callable
                progress.update(convert, advance=1, refresh=True)
                del r

    out_dstore.close()
    dvs_util.print_colour(
        f"Successfully created '{out_dstore.source!s}'",
        "green",
    )


@main.command(**_click_command_opts)
@_seqfile
@_outpath
@click.option(
    "-z",
    "--min_size",
    default=7,
    type=int,
    help="minimum size of divergent set",
)
@click.option(
    "-zp",
    "--max_size",
    default=None,
    type=int,
    help="maximum size of divergent set",
)
@_k
@click.option(
    "-st",
    "--stat",
    type=click.Choice(["stdev", "cov"]),
    default="stdev",
    help="statistic to maximise",
)
@_seed
@_include
@_numprocs
@_limit
@click.option(
    "-T",
    "--test_run",
    is_flag=True,
    help="reduce number of paths and size of query seqs",
)
@_verbose
@_hide_progress
def max(  # noqa: A001
    seqfile: Path,
    outpath: Path,
    min_size: int,
    max_size: int | None,
    stat: str,
    seed: int,
    include: list[str] | None,
    k: int,
    numprocs: int,
    limit: int | None,
    test_run: bool,
    verbose: int,
    hide_progress: bool,
) -> None:
    """Identify the seqs that maximise average delta JSD"""
    if max_size is not None and min_size > max_size:
        dvs_util.print_colour(f"{min_size=} cannot be greater than {max_size=}", "red")
        sys.exit(1)

    if seqfile.suffix != ".dvseqs":
        dvs_util.print_colour(
            "Sequence data needs to be preprocessed, use 'dvs prep'",
            "red",
        )
        sys.exit(1)

    seqids = dvs_data_store.get_seqids_from_store(seqfile)
    if len(seqids) < min_size:
        msg = f"Num seqs in {seqfile}={len(seqids)} < {min_size=}. Nothing to do!"
        dvs_util.print_colour(msg, "red")
        sys.exit(1)

    if include and not set(include) <= set(seqids):
        dvs_util.print_colour(
            f"provided {include=} not in the sequence data",
            "red",
        )
        sys.exit(1)

    if verbose:
        dvs_util.print_colour(f"Using random seed: {seed}", "blue")

    rng = numpy.random.default_rng(seed=seed)
    rng.shuffle(seqids)

    limit = min_size + 1 if test_run else limit
    if limit is not None:
        seqids = seqids[:limit]

    app = dvs_records.select_max(
        seq_store=seqfile,
        k=k,
        min_size=min_size,
        max_size=max_size,
        stat=stat,
        limit=limit,
        verbose=verbose,
    )
    # turn off pylint check, since the function is made into a class
    finalise = dvs_records.select_final_max(  # pylint: disable=no-value-for-parameter
        stat=stat,
        min_size=min_size,
        verbose=verbose,
    )
    result = dvs_records.apply_app(
        app=app,
        seqids=seqids,
        numprocs=numprocs,
        verbose=verbose,
        hide_progress=hide_progress,
        finalise=finalise,
    )

    # user requested inclusions are added to the selected divergent set
    if include:
        include_records = dvs_records.records_from_seq_store(
            seq_store=seqfile,
            seq_names=include,
            k=k,
            limit=None,
        )
        result = dvs_records.SummedRecords.from_records(
            result.all_records() + include_records,
        )

    outpath.parent.mkdir(parents=True, exist_ok=True)
    table = result.to_table()
    table.write(outpath)
    dvs_util.print_colour(
        f"{table.shape[0]} divergent sequences IDs written to '{outpath!s}'",
        "green",
    )


@main.command(**_click_command_opts)
@_seqfile
@_outpath
@click.option(
    "-n",
    "--number",
    type=int,
    help="number of seqs in divergent set",
    required=True,
)
@_k
@_seed
@_include
@_numprocs
@_limit
@_verbose
@_hide_progress
def nmost(
    seqfile: Path,
    outpath: Path,
    number: int,
    k: int,
    include: list[str] | None,
    seed: int,
    numprocs: int,
    limit: int | None,
    verbose: int,
    hide_progress: bool,
) -> None:
    """Identify n seqs that maximise average delta JSD"""

    if seqfile.suffix != ".dvseqs":
        dvs_util.print_colour(
            "Sequence data needs to be preprocessed, use 'dvs prep'",
            "red",
        )
        sys.exit(1)

    seqids = dvs_data_store.get_seqids_from_store(seqfile)
    if len(seqids) < number:
        msg = f"Num seqs in {seqfile}={len(seqids)} < {number=}. Nothing to do!"
        dvs_util.print_colour(msg, "red")
        sys.exit(1)

    if include and not set(include) <= set(seqids):
        dvs_util.print_colour(
            f"provided {include=} not in the sequence data",
            "red",
        )
        sys.exit(1)

    if verbose:
        dvs_util.print_colour(f"Using random seed: {seed}", "blue")

    rng = numpy.random.default_rng(seed=seed)
    rng.shuffle(seqids)
    if limit is not None:
        seqids = seqids[:limit]

    app = dvs_records.select_nmost(
        seq_store=seqfile,
        n=number,
        k=k,
        limit=limit,
        verbose=verbose,
    )
    result = dvs_records.apply_app(
        app=app,
        seqids=seqids,
        numprocs=numprocs,
        verbose=verbose,
        hide_progress=hide_progress,
        finalise=dvs_records.dvs_final_nmost(),  # pylint: disable=no-value-for-parameter
    )
    # user requested inclusions are added to the selected divergent set
    if include:
        include_records = dvs_records.records_from_seq_store(
            seq_store=seqfile,
            seq_names=include,
            k=k,
            limit=None,
        )
        result = dvs_records.SummedRecords.from_records(
            result.all_records() + include_records,
        )

    outpath.parent.mkdir(parents=True, exist_ok=True)
    table = result.to_table()
    table.write(outpath)
    dvs_util.print_colour(
        f"{table.shape[0]} divergent sequences IDs written to '{outpath!s}'",
        "green",
    )


@main.command(**_click_command_opts)
@_seqfile
@_outpath
@_moltype
@_k
@click.option(
    "--sketch-size",
    type=int,
    default=None,
    help="sketch size for mash distance, e.g. 3000",
)
@click.option(
    "-d",
    "--distance",
    type=click.Choice(["mash", "euclidean"]),
    default="mash",
    help="distance measure for tree construction",
)
@click.option(
    "-c",
    "--canonical-kmers",
    is_flag=True,
    default=False,
    help="consider kmers identical to their reverse complement",
)
@_seed
@_limit
@_numprocs
@_hide_progress
@_verbose
def ctree(
    seqfile: Path,
    outpath: Path,
    moltype: str,
    k: int,
    sketch_size: int | None,
    distance: str,
    canonical_kmers: bool,
    seed: int,
    limit: int | None,
    numprocs: int,
    hide_progress: bool,
    verbose: int,
):
    """Quickly compute a cluster tree based on kmers for a collection of sequences."""
    if seqfile.suffix != ".dvseqs":
        dvs_util.print_colour(
            "Sequence data needs to be preprocessed, use 'dvs prep'",
            "red",
        )
        sys.exit(1)

    if outpath is None:
        dvs_util.print_colour(
            "Specify a file name to write the newick tree to.",
            "red",
        )
        sys.exit(1)

    if sketch_size is None and distance == "mash":
        dvs_util.print_colour(
            "Sketch size must be specified for mash distance.",
            "red",
        )
        sys.exit(1)

    seqids = dvs_data_store.get_seqids_from_store(seqfile)
    if verbose:
        dvs_util.print_colour(f"Using random seed: {seed}", "blue")

    rng = numpy.random.default_rng(seed=seed)
    rng.shuffle(seqids)
    if limit is not None:
        seqids = seqids[:limit]

    app = dvs_cluster.dvs_cli_par_ctree(
        seq_store=seqfile,
        limit=limit,
        k=k,
        sketch_size=sketch_size,
        moltype=moltype,
        distance_mode=distance,
        mash_canonical_kmers=canonical_kmers,
        max_workers=numprocs,
        parallel=numprocs > 1,
        show_progress=not hide_progress,
    )
    tree = app(seqids)  # pylint: disable=not-callable
    if not tree:
        dvs_util.print_colour(tree, "red")
        sys.exit(1)

    tree.write(outpath)
    dvs_util.print_colour(f"Newick tree written to '{outpath!s}'", "green")


if __name__ == "__main__":
    main()



import functools
import string
import typing
from pathlib import Path

from attrs import define
from cogent3.app import typing as c3_types
from cogent3.app.composable import (
    LOADER,
    NON_COMPOSABLE,
    WRITER,
    define_app,
    source_proxy,
)
from cogent3.app.data_store import (
    OVERWRITE,
    DataMember,
    DataStoreABC,
    DataStoreDirectory,
    Mode,
)
from cogent3.core import new_alphabet
from cogent3.format import fasta as format_fasta
from cogent3.parse import fasta, genbank

from diverse_seq import util as dvs_utils
from diverse_seq.data_store import HDF5DataStore
from diverse_seq.record import SeqArray

converter_fasta = new_alphabet.convert_alphabet(
    string.ascii_lowercase.encode("utf8"),
    string.ascii_uppercase.encode("utf8"),
    delete=b"\n\r\t- ",
)

converter_genbank = new_alphabet.convert_alphabet(
    string.ascii_lowercase.encode("utf8"),
    string.ascii_uppercase.encode("utf8"),
    delete=b"\n\r\t- 0123456789",
)


def _label_func(label):
    return label.split()[0]


def _label_from_filename(path):
    return Path(path).stem.split(".")[0]


@define
class filename_seqname:
    source: str
    name: str


def get_format_parser(seq_path, seq_format):
    return (
        fasta.iter_fasta_records(seq_path, converter=converter_fasta)
        if seq_format == "fasta"
        else genbank.iter_genbank_records(
            seq_path,
            converter=converter_genbank,
            convert_features=None,
        )
    )


@define_app(app_type=LOADER)
class dvs_load_seqs:
    """Load and proprocess sequences from seq datastore"""

    def __init__(self, moltype: str = "dna", seq_format: str = "fasta"):
        """load fasta sequences from a data store

        Parameters
        ----------
        moltype
            molecular type

        Notes
        -----
        Assumes each fasta file contains a single sequence so only takes the first
        """
        self.moltype = moltype
        self.str2arr = dvs_utils.str2arr(moltype=self.moltype)
        self.seq_format = seq_format

    def main(self, data_member: DataMember) -> SeqArray:
        seq_path = Path(data_member.data_store.source) / data_member.unique_id
        parser = get_format_parser(seq_path, self.seq_format)
        seqs = [s for _, s, *_ in parser]
        return SeqArray(
            seqid=data_member.unique_id,
            data=self.str2arr(b"-".join(seqs).decode("utf8")),  # pylint: disable=not-callable
            moltype=self.moltype,
            source=data_member.data_store.source,
        )


@functools.singledispatch
def get_unique_id(val: typing.Any) -> str | None:
    for attr in ("unique_id", "seqid", "name", "source"):
        if result := getattr(val, attr, None):
            return get_unique_id(result)
    return None


@get_unique_id.register
def _(val: Path) -> str | None:
    # have to intercept Path objects as they have a name attribute
    return val.with_suffix("").name


@get_unique_id.register
def _(val: str) -> str | None:
    # have to intercept Path objects as they have a name attribute
    return get_unique_id(Path(val))


@get_unique_id.register
def _(val: SeqArray) -> str:
    return get_unique_id(val.seqid)


@define_app(app_type=WRITER)
class dvs_write_seqs:
    """Write seqs as numpy arrays to a HDF5DataStore"""

    def __init__(
        self,
        data_store: HDF5DataStore,
        id_from_source: callable = get_unique_id,
    ):
        self.data_store = data_store
        self.id_from_source = id_from_source

    def main(
        self,
        data: SeqArray,
        identifier: str | None = None,
    ) -> c3_types.IdentifierType:
        data = data.obj if isinstance(data, source_proxy) else data
        unique_id = identifier or self.id_from_source(data)
        return self.data_store.write(
            unique_id=unique_id,
            data=data.data,
            moltype=data.moltype,
            source=str(data.source),
        )


@define_app(app_type=NON_COMPOSABLE)
class dvs_file_to_dir:
    """Convert a single sequence file into a directory data store"""

    def __init__(
        self,
        dest: c3_types.IdentifierType | None = None,
        seq_format: str = "fasta",
        limit: int | None = None,
        mode: str | Mode = OVERWRITE,
    ):
        """
        Parameters
        ----------
        dest
            name for the output directory, defaults to name of the input file
        limit
            the number of sequences to be written defaults to all
        mode
            directory creation mode, by default OVERWRITE
        """
        self.dest = dest
        self.limit = limit
        self.mode = mode
        self.seq_format = seq_format

    def main(self, seq_path: c3_types.IdentifierType) -> DataStoreABC:
        outpath = Path(self.dest) if self.dest else Path(seq_path).with_suffix("")
        outpath.mkdir(parents=True, exist_ok=True)
        out_dstore = DataStoreDirectory(
            source=outpath,
            mode=self.mode,
            suffix=".fa",
            limit=self.limit,
        )

        parser = get_format_parser(seq_path, self.seq_format)
        seqs = {n: seq.replace(b"-", b"") for n, seq, *_ in parser}

        for seq_id, seq_data in seqs.items():
            fasta_seq_data = format_fasta.seqs_to_fasta(
                {seq_id: seq_data.decode("utf8")},
                block_size=1_000_000_000,
            )
            out_dstore.write(unique_id=seq_id, data=fasta_seq_data)

        return out_dstore



import pytest
from cogent3 import (
    get_app,
    load_aligned_seqs,
    load_unaligned_seqs,
    make_unaligned_seqs,
    open_data_store,
)
from cogent3.maths.measure import jsd
from numpy.testing import assert_allclose

from diverse_seq import records as dvs_records
from diverse_seq.record import KmerSeq, kmer_counts
from diverse_seq.util import str2arr


@pytest.fixture
def seqcoll():
    data = {"a": "AAAA", "b": "AAAA", "c": "TTTT", "d": "ACGT"}
    return make_unaligned_seqs(data, moltype="dna")


def _get_kfreqs_per_seq(seqs, k=1):
    app = str2arr()
    result = {}
    for seq in seqs.seqs:
        arr = app(str(seq))  # pylint: disable=not-callable
        freqs = kmer_counts(arr, 4, k)
        result[seq.name] = freqs
    return result


def _make_records(kcounts, seqcoll):
    return [
        KmerSeq(kcounts=kcounts[s.name], name=s.name)
        for s in seqcoll.seqs
        if s.name in kcounts
    ]


@pytest.mark.parametrize("k", (1, 2))
def test_total_jsd(seqcoll, k):
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = [
        KmerSeq(kcounts=kcounts[s.name], name=f"{s.name}-{k}") for s in seqcoll.seqs
    ]
    sr = dvs_records.SummedRecords.from_records(records)
    freqs = {n: v.astype(float) / v.sum() for n, v in kcounts.items()}
    expect = jsd(*list(freqs.values()))
    assert_allclose(sr.total_jsd, expect)


def test_add(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    sr = dvs_records.SummedRecords.from_records(records[:-1])
    orig = id(sr)
    sr = sr + records[-1]
    assert id(sr) != orig
    freqs = {n: v.astype(float) / v.sum() for n, v in kcounts.items()}
    expect = jsd(*list(freqs.values()))
    assert_allclose(sr.total_jsd, expect)


def test_sub(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    sr = dvs_records.SummedRecords.from_records(records)
    assert sr.size == 4
    orig = id(sr)
    sr = sr - records[-1]
    assert id(sr) != orig
    assert sr.size == 3
    freqs = {n: v.astype(float) / v.sum() for n, v in kcounts.items()}
    expect = jsd(*[freqs[name] for name in sr.iter_record_names()])
    assert_allclose(sr.total_jsd, expect)


@pytest.mark.parametrize(
    "exclude,expect",
    (("a", False), ("b", False), ("c", True), ("d", True)),
)
def test_increases_jsd(seqcoll, exclude, expect):
    kcounts = _get_kfreqs_per_seq(seqcoll, k=1)
    records = {r.name: r for r in _make_records(kcounts, seqcoll)}
    excluded = records.pop(exclude)
    sr = dvs_records.SummedRecords.from_records(list(records.values()))
    assert sr.increases_jsd(excluded) == expect


def test_mean_delta_jsd(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = [KmerSeq(kcounts=kcounts[s.name], name=s.name) for s in seqcoll.seqs]
    sr_with_a = dvs_records.SummedRecords.from_records(records)
    sr_without_a = dvs_records.SummedRecords.from_records(
        [r for r in records if r.name != "a"],
    )
    assert sr_without_a.mean_delta_jsd > sr_with_a.mean_delta_jsd


def test_replaced_lowest(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    sr = dvs_records.SummedRecords.from_records(records[:-1])
    lowest = sr.lowest
    nsr = sr.replaced_lowest(records[-1])

    assert nsr is not sr
    assert sr.size == nsr.size == len(records) - 1
    # make sure previous lowest not present at all
    assert nsr.lowest is not lowest
    for r in nsr.records:
        assert r is not lowest
    # make sure new record is present
    assert any(r is records[-1] for r in nsr.records + [nsr.lowest])


def test_max_divergent(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    got = dvs_records.max_divergent(records, min_size=2, max_size=2, stat="stdev")
    assert got.size == 2


def test_most_divergent(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    got = dvs_records.most_divergent(records, size=3)
    assert got.size == 3
    assert got.to_table().shape == (3, 2)


def test_all_records(seqcoll):
    k = 1
    kcounts = _get_kfreqs_per_seq(seqcoll, k=k)
    records = _make_records(kcounts, seqcoll)
    got = dvs_records.most_divergent(records, size=3).all_records()
    assert len(got) == 3
    assert isinstance(got[0], KmerSeq)


@pytest.fixture(scope="session")
def brca1_coll(DATA_DIR):
    return load_unaligned_seqs(DATA_DIR / "brca1.fasta", moltype="dna").degap()


@pytest.fixture(scope="session")
def brca1_alignment(DATA_DIR):
    return load_aligned_seqs(DATA_DIR / "brca1.fasta", moltype="dna")


def test_merge_summed_records(DATA_DIR, brca1_coll):
    path = DATA_DIR / "brca1.dvseqs"
    names = brca1_coll.names
    app = dvs_records.select_nmost(seq_store=path, n=5, k=1)
    sr1 = app(names[:10])  # pylint: disable=not-callable
    sr2 = app(names[10:20])  # pylint: disable=not-callable
    rnames1 = sr1.record_names
    rnames2 = sr2.record_names
    assert set(rnames1) != set(rnames2)
    got = dvs_records.dvs_final_nmost()([sr1, sr2])
    assert len(got.record_names) == 5


def test_dvs_select_max(brca1_coll):
    app = dvs_records.dvs_max(k=1, min_size=2, max_size=5)
    got = app(brca1_coll)  # pylint: disable=not-callable
    assert 2 <= got.num_seqs <= 5
    app = dvs_records.dvs_max(k=1, min_size=2, max_size=5, seed=123)
    got = app(brca1_coll)  # pylint: disable=not-callable
    assert 2 <= got.num_seqs <= 5


@pytest.mark.parametrize("include", ("Human", ["Human"], ["Human", "Mouse"]))
def test_dvs_select_max_include(brca1_coll, include):
    app = dvs_records.dvs_max(k=1, min_size=2, max_size=5, include=include)
    got = app(brca1_coll)  # pylint: disable=not-callable
    include = {include} if isinstance(include, str) else set(include)
    assert include <= set(got.names)


def test_dvs_select_nmost(brca1_coll):
    app = dvs_records.dvs_nmost(k=1, n=5)
    got = app(brca1_coll)  # pylint: disable=not-callable
    assert got.num_seqs == 5
    # try setting a seed
    app = dvs_records.dvs_nmost(k=1, n=5, seed=123)
    got = app(brca1_coll)  # pylint: disable=not-callable
    assert got.num_seqs == 5


@pytest.mark.parametrize("include", ("Human", ["Human"], ["Human", "Mouse"]))
def test_dvs_select_nmost_keep(brca1_coll, include):
    app = dvs_records.dvs_nmost(k=1, n=5, include=include)
    got = app(brca1_coll)  # pylint: disable=not-callable
    include = {include} if isinstance(include, str) else set(include)
    assert include <= set(got.names)


@pytest.mark.parametrize("app_name", ("dvs_nmost", "dvs_max"))
def test_serialisable_nmost(brca1_coll, tmp_path, app_name):
    brca1_coll.info.source = "blah.fa"
    outstore = open_data_store(tmp_path / "data.sqlitedb", mode="w")
    select = get_app(app_name, k=2)
    writer = get_app("write_db", data_store=outstore)
    app = select + writer
    got = app(brca1_coll)  # pylint: disable=not-callable
    assert len(outstore.completed) == 1


@pytest.mark.parametrize("app_name", ("dvs_nmost", "dvs_max"))
@pytest.mark.parametrize("aligned", (False, True))
def test_select_return_type(brca1_alignment, app_name, aligned):
    coll = brca1_alignment if aligned else brca1_alignment.degap()
    select = get_app(app_name, k=2)
    got = select(coll)  # pylint: disable=not-callable
    assert isinstance(got, coll.__class__)



from collections import Counter
from itertools import product
from pathlib import Path

import numpy
import pytest
from cogent3 import get_moltype, make_seq
from numpy import (
    array,
    nextafter,
    ravel_multi_index,
    uint16,
    uint64,
    unravel_index,
    zeros,
)
from numpy.testing import assert_allclose

from diverse_seq.record import (
    KmerSeq,
    SeqArray,
    _gettype,
    coord_conversion_coeffs,
    coord_to_index,
    index_to_coord,
    indices_to_seqs,
    kmer_counts,
    lazy_kmers,
    make_kmerseq,
    member_to_kmerseq,
    seqarray_to_kmerseq,
    vector,
)
from diverse_seq.util import str2arr

DATADIR = Path(__file__).parent / "data"


seq5 = make_seq("ACGCG", name="null", moltype="dna")
seq0 = make_seq("", name="null", moltype="dna")
seq_nameless = make_seq("ACGTGTT", moltype="dna")
seq4eq = make_seq("ACGT", name="null", moltype="dna")
seq4one = make_seq("AAAA", name="null", moltype="dna")


@pytest.fixture
def seqarray():
    return SeqArray(
        seqid="seq1",
        data=numpy.array([0, 1, 2, 3]),
        moltype="dna",
        source="seq1_source",
    )


@pytest.mark.parametrize("seq,entropy", ((seq4eq, 2.0), (seq4one, 0.0)))
def test_seqrecord_entropy(seq, entropy):
    arr = str2arr()(str(seq))  # pylint: disable=not-callable
    kcounts = kmer_counts(arr, 4, 1)
    sr = KmerSeq(kcounts=kcounts, name=seq.name)
    assert sr.entropy == entropy


@pytest.mark.parametrize("name", (1, 1.0))
def test_seqrecord_invalid_types(name):
    kcounts = array([1, 2, 3])
    with pytest.raises(TypeError):
        KmerSeq(kcounts=kcounts, name=name)


@pytest.mark.parametrize("kcounts", ((0, 1, 2), [0, 1, 2]))
def test_seqrecord_invalid_kcounts(kcounts):
    with pytest.raises(TypeError):
        KmerSeq(kcounts=kcounts, name="n1")


def test_seqrecord_compare():
    arr = str2arr()(str(seq5))  # pylint: disable=not-callable
    kcounts = kmer_counts(arr, 4, 2)
    kwargs = dict(kcounts=kcounts)
    sr1 = KmerSeq(name="n1", **kwargs)
    sr1.delta_jsd = 1.0
    sr2 = KmerSeq(name="n2", **kwargs)
    sr2.delta_jsd = 2.0
    sr3 = KmerSeq(name="n3", **kwargs)
    sr3.delta_jsd = 34.0

    rec = sorted((sr3, sr1, sr2))
    assert rec[0].delta_jsd == 1 and rec[-1].delta_jsd == 34


def test_sparse_vector_create():
    from numpy import array

    data = {2: 3, 3: 9}
    # to constructor
    v1 = vector(data=data, vector_length=2**2, dtype=int)
    expect = array([0, 0, 3, 9])
    assert_allclose(v1.data, expect)

    # or via set item individually
    v2 = vector(vector_length=2**2)
    for index, count in data.items():
        v2[index] = count

    assert (v1.data == v2.data).all()


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_add_vector(cast):
    # adds two vectors
    data = {2: 3, 3: 9}
    v1 = vector(data=data, vector_length=2**2, dtype=cast)

    # add to zero vector
    v0 = vector(vector_length=2**2, dtype=cast)
    v_ = v1 + v0
    assert (v_.data == v1.data).all()
    assert v_.data is not v1.data

    # add to self
    v3 = v1 + v1
    assert (v3.data == v1.data * 2).all()

    # add in-place
    v1 += v1
    assert (v1.data == v3.data).all()


@pytest.mark.parametrize("zero", (0, 0.0, nextafter(0.0, 1.0) / 2))
def test_sparse_vector_add_zero(zero):
    # does not include
    expect = {2: 3.0, 3: 9.0}
    expect = numpy.zeros(4, dtype=float)
    expect[2] = 3.0
    expect[3] = 9.0
    data = expect[:]
    data[1] = 0.0
    v1 = vector(vector_length=2**2, data=data, dtype=float)
    assert (v1.data == expect).all()


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_add_scalar(cast):
    # adds two vectors
    data = {2: 3, 3: 9}
    v1 = vector(data=data, vector_length=2**2, dtype=cast)

    # add to zero vector
    v_ = v1 + 0
    assert (v_.data == v1.data).all()
    assert v_.data is not v1.data

    # add in place
    expect = v1.data[:]
    expect += 5
    v1 += 5
    assert (v1.data == expect).all()


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_sub_vector(cast):
    # sub two vectors
    data = {2: 3, 3: 9}
    v1 = vector(data=data, vector_length=2**2, dtype=cast)
    # sub self
    v3 = v1 - v1
    assert_allclose(v3.data, 0)

    data2 = {2: 3, 3: 10}
    v2 = vector(data=data2, vector_length=2**2, dtype=cast)
    v3 = v2 - v1
    expect = numpy.zeros(4, dtype=cast)
    expect[3] = cast(1)
    assert (v3.data == expect).all()

    # sub in-place
    orig = v1
    expect = v1.data - v2.data
    v1 -= v2
    assert v1 is orig
    assert (v1.data == expect).all()

    # negative allowed
    data = {2: 3, 3: 9}
    v1 = vector(data=data, vector_length=2**2, dtype=cast)
    data2 = {2: 6, 3: 10}
    v2 = vector(data=data2, vector_length=2**2, dtype=cast)
    v3 = v1 - v2
    assert v3[2] == -3


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_sub_scalar(cast):
    # subtracts two vectors
    data = {2: 3, 3: 9}
    v1 = vector(data=data, vector_length=2**2, dtype=cast)

    # subtract zero
    v_ = v1 - 0
    assert (v_.data == v1.data).all()
    assert v_.data is not v1.data

    # subtract in place
    expect = v1.data - 2
    v1 -= 2
    assert (v1.data == expect).all()


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_sub_elementwise(cast):
    data = {2: 3, 3: 9}
    v2 = vector(data=data, vector_length=2**2, dtype=cast)
    v2[1] -= 99
    assert v2[1] == -99
    text = "float" if cast == float else "int"
    assert v2[1].dtype.name.startswith(text)


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_elementwise(cast):
    data = {2: 3, 3: 9}
    v2 = vector(data=data, vector_length=2**2, dtype=cast)
    v2[1] += 99
    assert v2[1] == 99

    text = "float" if cast == float else "int"
    assert v2[1].dtype.name.startswith(text)


def test_sparse_vector_sum():
    sv = vector(vector_length=2**2, dtype=int)
    assert sv.sum() == 0


def test_sparse_vector_iter_nonzero():
    data = {3: 9, 2: 3}
    sv = vector(data=data, vector_length=4**2, dtype=int)
    got = list(sv.iter_nonzero())
    assert got == [3, 9]


def test_sv_iter():
    sv = vector(data={2: 1, 1: 1, 3: 1, 0: 1}, vector_length=2**2, dtype=int)
    sv /= sv.sum()
    got = list(sv.iter_nonzero())
    assert_allclose(got, 0.25)


def test_sv_pickling():
    import pickle

    o = vector(data={2: 1, 1: 1, 3: 1, 0: 1}, vector_length=2**2, dtype=int)
    p = pickle.dumps(o)
    u = pickle.loads(p)
    assert str(u) == str(o)


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_div_vector(cast):
    # adds two vectors
    data1 = {2: 6, 3: 18}
    v1 = vector(data=data1, vector_length=2**2, dtype=cast)

    # factored by 3
    data2 = {2: 3, 3: 3}
    v2 = vector(data=data2, vector_length=2**2, dtype=cast)
    expect = numpy.nan_to_num(v1.data / v2.data, nan=0.0)
    v3 = v1 / v2
    assert (v3.data == expect).all()

    # different factors
    data2 = {2: 3, 3: 6}
    v2 = vector(data=data2, vector_length=2**2, dtype=cast)
    v3 = v1 / v2
    expect = numpy.nan_to_num(v1.data / v2.data, nan=0.0)
    assert (v3.data == expect).all()

    # div in-place
    orig = v1
    v1 /= v2
    assert v1 is orig
    assert (v1.data == expect).all()


@pytest.mark.parametrize("cast", (float, int))
def test_sparse_vector_div_scalar(cast):
    # adds two vectors
    data1 = {2: 6, 3: 18}
    v1 = vector(data=data1, vector_length=2**2, dtype=cast)

    # factored by 3
    v2 = v1 / 3
    expect = numpy.nan_to_num(v1.data / 3, nan=0.0)
    assert (v2.data == expect).all()
    assert v2 is not v1

    # div in-place
    v1 /= 3
    assert (v1.data == expect).all()


@pytest.mark.parametrize("num_states,ndim", ((2, 1), (4, 2), (4, 4)))
def test_interconversion_of_coords_indices(num_states, ndim):
    # make sure match numpy functions
    coeffs = array(coord_conversion_coeffs(num_states, ndim))
    for coord in product(*(list(range(num_states)),) * ndim):
        coord = array(coord, dtype=uint64)
        nidx = ravel_multi_index(coord, dims=(num_states,) * ndim)
        idx = coord_to_index(coord, coeffs)
        assert idx == nidx
        ncoord = unravel_index(nidx, shape=(num_states,) * ndim)
        got = index_to_coord(idx, coeffs)
        assert (got == ncoord).all()
        assert (got == coord).all()


def test_coord2index_fail():
    coord = array((0, 1), dtype=uint64)
    coeffs = array(coord_conversion_coeffs(2, 4))
    # dimension of coords inconsistent with coeffs
    with pytest.raises(ValueError):
        coord_to_index(coord, coeffs)


_seqs = ("ACGGCGGTGCA", "ACGGNGGTGCA", "ANGGCGGTGNA")
_ks = (1, 2, 3)


def test_indices_to_seqs():
    indices = array([0, 8], dtype=uint16)
    states = b"TCAG"
    result = indices_to_seqs(indices, states, 2)
    assert result == ["TT", "AT"]
    indices = array([0, 16], dtype=uint16)
    with pytest.raises(IndexError):
        # 16 is outside range
        indices_to_seqs(indices, states, 2)


@pytest.mark.parametrize("seq,k", tuple(product(_seqs, _ks)))
def test_kmer_freqs(seq, k):
    seq = make_seq(seq, moltype="dna")
    states = (list(seq.moltype),) * k
    states = ["".join(v) for v in product(*states)]
    counts = Counter(seq.iter_kmers(k))
    expect = zeros(len(states), dtype=int)
    for kmer, v in counts.items():
        if kmer not in states:
            continue
        expect[states.index(kmer)] = v

    seq2array = str2arr()
    arr = seq2array(str(seq))  # pylint: disable=not-callable
    got = kmer_counts(arr, 4, k)
    assert (got == expect).all()


def test_composable():
    from cogent3.app.composable import define_app

    @define_app
    def matched_sr(records: list[KmerSeq]) -> bool:
        return True

    sr = KmerSeq(kcounts=numpy.array([1, 2, 3, 4], dtype=int), name="a")
    app = matched_sr()  # pylint: disable=no-value-for-parameter
    got = app([sr])  # pylint: disable=not-callable
    assert got


dtype_name = lambda x: numpy.dtype(x).name


@pytest.mark.parametrize(
    "dtype",
    ("int", "float", dtype_name(numpy.int32), dtype_name(numpy.float64)),
)
def test__gettype(dtype):
    if dtype[-1].isdigit():
        expect = getattr(numpy, dtype)
    else:
        expect = int if dtype == "int" else float
    got = _gettype(dtype)
    assert got == expect


@pytest.mark.parametrize("k", (1, 2, 3))
def test_seqarray_to_record(seqarray, k):
    s2r = seqarray_to_kmerseq(k=k, moltype="dna")
    rec = s2r(seqarray)  # pylint: disable=not-callable

    assert rec.name == "seq1"
    rec.kfreqs.sum() == len(seqarray) - k + 1


@pytest.fixture(scope="function")
def dstore():
    from diverse_seq.data_store import HDF5DataStore

    store = HDF5DataStore("test.h5", in_memory=True)
    yield store
    store.close()


def test_lazy_kmers(dstore, seqarray):
    member = dstore.write(unique_id="seq1", data=seqarray.data)

    lazy = lazy_kmers(data=member, k=2, moltype="dna")
    s2k = seqarray_to_kmerseq(k=2, moltype="dna")
    expect = s2k(seqarray)  # pylint: disable=not-callable
    assert_allclose(numpy.array(lazy), expect.kcounts)


def test_member_to_vector(dstore, seqarray):
    member = dstore.write(unique_id="seq1", data=seqarray.data)
    lazy = lazy_kmers(data=member, k=2, moltype="dna")
    vec = vector(data=lazy, vector_length=lazy.num_states)
    assert_allclose(numpy.array(vec), numpy.array(lazy))


def test_member_to_kmerseq(dstore, seqarray):
    member = dstore.write(unique_id="seq1", data=seqarray.data)
    new_app = member_to_kmerseq(k=2, moltype="dna")
    got = new_app(member)  # pylint: disable=not-callable
    assert isinstance(got, KmerSeq)
    old_app = seqarray_to_kmerseq(k=2, moltype="dna")
    expect = old_app(seqarray)  # pylint: disable=not-callable
    assert_allclose(got.kcounts, expect.kcounts)


@pytest.mark.parametrize("new_seq", (True, False))
def test_make_kmerseq(new_seq):
    from numpy import min_scalar_type

    dtype = min_scalar_type(16)
    raw = "ACGT"
    name = "seq1"
    seq = make_seq(seq=raw, name=name, moltype="dna", new_type=new_seq)
    seq.info.source = "source1"
    kseq = make_kmerseq(seq, dtype=dtype, k=2, moltype="dna")
    alpha = get_moltype("dna", new_type=True).alphabet.get_kmer_alphabet(2)
    expect_kmers = alpha.to_indices(raw, independent_kmer=False)
    expect = zeros(len(alpha), dtype=dtype)
    expect[expect_kmers] = 1
    assert_allclose(array(kseq.kcounts), expect)



"""Apps and methods used to compute kmer cluster trees for sequences."""

import multiprocessing
from collections.abc import Sequence
from contextlib import nullcontext
from pathlib import Path
from typing import Literal

import cogent3.app.typing as c3_types
import numpy
from cogent3 import PhyloNode, make_tree
from cogent3.app.composable import AppType, define_app
from loky import as_completed, get_reusable_executor
from rich.progress import Progress
from scipy.sparse import dok_matrix
from sklearn.cluster import AgglomerativeClustering

from diverse_seq import data_store as dvs_data_store
from diverse_seq.distance import (
    BottomSketch,
    euclidean_distance,
    euclidean_distances,
    mash_distance,
    mash_distances,
    mash_sketch,
)
from diverse_seq.record import (
    KmerSeq,
    SeqArray,
    _get_canonical_states,
    make_kmerseq,
    seq_to_seqarray,
)


class ClusterTreeBase:
    def __init__(
        self,
        *,
        k: int = 16,
        sketch_size: int | None = None,
        moltype: str = "dna",
        distance_mode: Literal["mash", "euclidean"] = "mash",
        mash_canonical_kmers: bool | None = None,
        show_progress: bool = False,
    ) -> None:
        """Initialise parameters for generating a kmer cluster tree.

        Parameters
        ----------
        k
            kmer size
        sketch_size
            size of sketches
        moltype
            seq collection molecular type
        distance_mode
            mash distance or euclidean distance between kmer freqs
        mash_canonical_kmers
            whether to use mash canonical kmers for mash distance
        show_progress
            whether to show progress bars

        Notes
        -----
        If mash_canonical_kmers is enabled when using the mash distance,
        kmers are considered identical to their reverse complement.

        References
        ----------
        .. [1] Ondov, B. D., Treangen, T. J., Melsted, P., Mallonee, A. B.,
           Bergman, N. H., Koren, S., & Phillippy, A. M. (2016).
           Mash: fast genome and metagenome distance estimation using MinHash.
           Genome biology, 17, 1-14.
        """
        if mash_canonical_kmers is None:
            mash_canonical_kmers = False

        if distance_mode not in ("mash", "euclidean"):
            msg = f"Unexpected distance {distance_mode!r}."
            raise ValueError(msg)

        if moltype not in ("dna", "rna") and mash_canonical_kmers:
            msg = "Canonical kmers only supported for dna/rna sequences."
            raise ValueError(msg)

        if distance_mode == "mash" and sketch_size is None:
            msg = "Expected sketch size for mash distance measure."
            raise ValueError(msg)

        self._moltype = moltype
        self._k = k
        self._num_states = len(_get_canonical_states(self._moltype))
        self._sketch_size = sketch_size
        self._distance_mode = distance_mode
        self._mash_canonical = mash_canonical_kmers
        self._progress = Progress(disable=not show_progress)

        self._s2a = seq_to_seqarray(moltype=moltype)


@define_app
class dvs_ctree(ClusterTreeBase):
    """Create a cluster tree from kmer distances."""

    def __init__(
        self,
        *,
        k: int = 12,
        sketch_size: int | None = 3_000,
        moltype: str = "dna",
        distance_mode: Literal["mash", "euclidean"] = "mash",
        mash_canonical_kmers: bool | None = None,
        show_progress: bool = False,
    ) -> None:
        """Initialise parameters for generating a kmer cluster tree.

        Parameters
        ----------
        k
            kmer size
        sketch_size
            size of sketches, only applies to mash distance
        moltype
            seq collection molecular type
        distance_mode
            mash distance or euclidean distance between kmer freqs
        mash_canonical_kmers
            whether to use mash canonical kmers for mash distance
        show_progress
            whether to show progress bars

        Notes
        -----
        This app is composable.

        If mash_canonical_kmers is enabled when using the mash distance,
        kmers are considered identical to their reverse complement.

        References
        ----------
        .. [1] Ondov, B. D., Treangen, T. J., Melsted, P., Mallonee, A. B.,
           Bergman, N. H., Koren, S., & Phillippy, A. M. (2016).
           Mash: fast genome and metagenome distance estimation using MinHash.
           Genome biology, 17, 1-14.
        """
        super().__init__(
            k=k,
            sketch_size=sketch_size,
            moltype=moltype,
            distance_mode=distance_mode,
            mash_canonical_kmers=mash_canonical_kmers,
            show_progress=show_progress,
        )

    def main(self, seqs: c3_types.SeqsCollectionType) -> PhyloNode:
        """Construct a cluster tree for a collection of sequences.

        Parameters
        ----------
        seqs
            Sequence collection to form cluster tree for.

        Returns
        -------
        PhyloNode
            a cluster tree.
        """
        seqs = seqs.degap()
        seq_names = seqs.names
        seq_arrays = [self._s2a(seqs.get_seq(name)) for name in seq_names]  # pylint: disable=not-callable

        with self._progress:
            if self._distance_mode == "mash":
                distances = mash_distances(
                    seq_arrays,
                    self._k,
                    self._sketch_size,
                    self._num_states,
                    mash_canonical=self._mash_canonical,
                    progress=self._progress,
                )
            elif self._distance_mode == "euclidean":
                distances = euclidean_distances(
                    seq_arrays,
                    self._k,
                    self._moltype,
                    progress=self._progress,
                )
            return make_cluster_tree(seq_names, distances, progress=self._progress)


def make_cluster_tree(
    seq_names: Sequence[str],
    pairwise_distances: numpy.ndarray,
    *,
    progress: Progress | None = None,
) -> PhyloNode:
    """Given pairwise distances between sequences, construct a cluster tree.

    Parameters
    ----------
    seq_names
        Names of sequences to cluster.
    pairwise_distances
        Pairwise distances between clusters.
    progress
        Progress bar, by default None.

    Returns
    -------
    PhyloNode
        The cluster tree.
    """
    if progress is None:
        progress = Progress(disable=True)

    clustering = AgglomerativeClustering(
        metric="precomputed",
        linkage="average",
    )

    tree_task = progress.add_task(
        "[green]Computing Tree",
        total=None,
    )

    clustering.fit(pairwise_distances)

    tree_dict = {i: seq_names[i] for i in range(len(seq_names))}
    node_index = len(seq_names)
    for left_index, right_index in clustering.children_:
        tree_dict[node_index] = (
            tree_dict.pop(left_index),
            tree_dict.pop(right_index),
        )
        node_index += 1
    # use string representation and then remove quotes
    treestring = str(tree_dict[node_index - 1]).replace("'", "")
    tree = make_tree(treestring=treestring, underscore_unmunge=True)
    progress.update(tree_task, completed=1, total=1)
    return tree


class DvsParCtreeMixin:
    def _mash_dist(self, seq_arrays: Sequence[SeqArray]) -> numpy.ndarray:
        """Calculates pairwise mash distances between sequences in parallel.

        Uses the number of processes specified in the constructor. If one process was
        specified runs in serial.

        Parameters
        ----------
        seq_arrays
            Sequence arrays.

        Returns
        -------
        numpy.ndarray
            Pairwise mash distances between sequences.
        """
        if self._numprocs == 1:
            return mash_distances(
                seq_arrays,
                self._k,
                self._sketch_size,
                self._num_states,
                mash_canonical=self._mash_canonical,
                progress=self._progress,
            )

        sketches = self.mash_sketches_parallel(seq_arrays)

        distance_task = self._progress.add_task(
            "[green]Computing Pairwise Distances",
            total=None,
        )

        # Compute distances in parallel
        num_jobs = self._numprocs
        futures = [
            self._executor.submit(
                compute_mash_chunk_distances,
                start,
                num_jobs,
                sketches,
                self._k,
                self._sketch_size,
            )
            for start in range(num_jobs)
        ]

        distances = numpy.zeros((len(sketches), len(sketches)))

        for future in futures:
            start_idx, distances_chunk = future.result()
            distances[start_idx::num_jobs] = distances_chunk[
                start_idx::num_jobs
            ].toarray()

        # Make lower triangular matrix symmetric
        distances = distances + distances.T - numpy.diag(distances.diagonal())

        self._progress.update(distance_task, completed=1, total=1)

        return distances

    def _euclidean_dist(
        self,
        seq_arrays: Sequence[SeqArray],
    ) -> numpy.ndarray:
        """Calculates pairwise euclidean distances between sequences in parallel.

        Uses the number of processes specified in the constructor. If one process was
        specified runs in serial.

        Parameters
        ----------
        seq_arrays
            Sequence arrays.

        Returns
        -------
        numpy.ndarray
            Pairwise euclidean distances between sequences.
        """
        if self._numprocs == 1:
            return euclidean_distances(
                seq_arrays,
                self._k,
                self._moltype,
                progress=self._progress,
            )

        distance_task = self._progress.add_task(
            "[green]Computing Pairwise Distances",
            total=None,
        )

        kmer_seqs = [
            make_kmerseq(
                seq,
                dtype=numpy.min_scalar_type(
                    len(_get_canonical_states(self._moltype)) ** self._k,
                ),
                k=self._k,
                moltype=self._moltype,
            )
            for seq in seq_arrays
        ]

        # Compute distances in parallel
        num_jobs = self._numprocs
        futures = [
            self._executor.submit(
                compute_euclidean_chunk_distances,
                start,
                num_jobs,
                kmer_seqs,
            )
            for start in range(num_jobs)
        ]

        distances = numpy.zeros((len(kmer_seqs), len(kmer_seqs)))

        for future in futures:
            start_idx, distances_chunk = future.result()
            distances[start_idx::num_jobs] = distances_chunk[
                start_idx::num_jobs
            ].toarray()

        # Make lower triangular matrix symmetric
        distances = distances + distances.T - numpy.diag(distances.diagonal())

        self._progress.update(distance_task, completed=1, total=1)
        return distances

    def mash_sketches_parallel(
        self,
        seq_arrays: Sequence[SeqArray],
    ) -> list[BottomSketch]:
        """Create sketch representations for a collection of sequence records in parallel.

        Parameters
        ----------
        seq_arrays
            Sequence arrays.

        Returns
        -------
        list[BottomSketch]
            Sketches for each sequence.
        """
        sketch_task = self._progress.add_task(
            "[green]Generating Sketches",
            total=len(seq_arrays),
        )

        bottom_sketches = [None for _ in range(len(seq_arrays))]

        # Compute sketches in parallel
        futures_to_idx = {
            self._executor.submit(
                mash_sketch,
                seq_array.data,
                self._k,
                self._sketch_size,
                self._num_states,
                mash_canonical=self._mash_canonical,
            ): i
            for i, seq_array in enumerate(seq_arrays)
        }

        for future in as_completed(list(futures_to_idx)):
            idx = futures_to_idx[future]
            bottom_sketches[idx] = future.result()

            self._progress.update(sketch_task, advance=1)

        return bottom_sketches


@define_app(app_type=AppType.NON_COMPOSABLE)
class dvs_par_ctree(ClusterTreeBase, DvsParCtreeMixin):
    """Create a cluster tree from kmer distances in parallel."""

    def __init__(
        self,
        *,
        k: int = 12,
        sketch_size: int | None = 3000,
        moltype: str = "dna",
        distance_mode: Literal["mash", "euclidean"] = "mash",
        mash_canonical_kmers: bool | None = None,
        show_progress: bool = False,
        max_workers: int | None = None,
        parallel: bool = True,
    ) -> None:
        """Initialise parameters for generating a kmer cluster tree.

        Parameters
        ----------
        k
            kmer size
        sketch_size
            size of sketches, only applies to mash distance
        moltype
            seq collection molecular type
        distance_mode
            mash distance or euclidean distance between kmer freqs
        mash_canonical_kmers
            whether to use mash canonical kmers for mash distance
        show_progress
            whether to show progress bars
        numprocs
            number of workers, defaults to running serial

        Notes
        -----
        This is app is not composable but can run in parallel. It is
        best suited to a single large sequence collection.

        If mash_canonical_kmers is enabled when using the mash distance,
        kmers are considered identical to their reverse complement.

        References
        ----------
        .. [1] Ondov, B. D., Treangen, T. J., Melsted, P., Mallonee, A. B.,
           Bergman, N. H., Koren, S., & Phillippy, A. M. (2016).
           Mash: fast genome and metagenome distance estimation using MinHash.
           Genome biology, 17, 1-14.
        """
        super().__init__(
            k=k,
            sketch_size=sketch_size,
            moltype=moltype,
            distance_mode=distance_mode,
            mash_canonical_kmers=mash_canonical_kmers,
            show_progress=show_progress,
        )

        if parallel:
            if max_workers is None:
                max_workers = multiprocessing.cpu_count()
            self._numprocs = min(max_workers, multiprocessing.cpu_count())
        else:
            self._numprocs = 1

        self._calc_dist = (
            self._mash_dist if distance_mode == "mash" else self._euclidean_dist
        )

    def main(self, seqs: c3_types.SeqsCollectionType) -> PhyloNode:
        """Construct a cluster tree for a collection of sequences.

        Parameters
        ----------
        seqs
            Sequence collection to form cluster tree for.

        Returns
        -------
        PhyloNode
            a cluster tree.
        """
        seqs = seqs.degap()
        self._executor = (
            get_reusable_executor(max_workers=self._numprocs)
            if self._numprocs != 1
            else nullcontext()
        )

        seq_names = seqs.names
        seq_arrays = [self._s2a(seqs.get_seq(name)) for name in seq_names]  # pylint: disable=not-callable

        with self._progress, self._executor:
            distances = self._calc_dist(seq_arrays)
            return make_cluster_tree(seq_names, distances, progress=self._progress)


@define_app(app_type=AppType.NON_COMPOSABLE)
class dvs_cli_par_ctree(ClusterTreeBase, DvsParCtreeMixin):
    """Create a cluster tree from kmer distances in parallel."""

    def __init__(
        self,
        *,
        seq_store: str | Path,
        limit: int | None = None,
        k: int = 12,
        sketch_size: int | None = 3000,
        moltype: str = "dna",
        distance_mode: Literal["mash", "euclidean"] = "mash",
        mash_canonical_kmers: bool | None = None,
        show_progress: bool = False,
        max_workers: int | None = None,
        parallel: bool = True,
    ) -> None:
        """Initialise parameters for generating a kmer cluster tree.

        Parameters
        ----------
        k
            kmer size
        sketch_size
            size of sketches, only applies to mash distance
        moltype
            seq collection molecular type
        distance_mode
            mash distance or euclidean distance between kmer freqs
        mash_canonical_kmers
            whether to use mash canonical kmers for mash distance
        show_progress
            whether to show progress bars
        numprocs
            number of workers, defaults to running serial

        Notes
        -----
        This is app is not composable but can run in parallel. It is
        best suited to a single large sequence collection.

        If mash_canonical_kmers is enabled when using the mash distance,
        kmers are considered identical to their reverse complement.

        References
        ----------
        .. [1] Ondov, B. D., Treangen, T. J., Melsted, P., Mallonee, A. B.,
           Bergman, N. H., Koren, S., & Phillippy, A. M. (2016).
           Mash: fast genome and metagenome distance estimation using MinHash.
           Genome biology, 17, 1-14.
        """
        super().__init__(
            k=k,
            sketch_size=sketch_size,
            moltype=moltype,
            distance_mode=distance_mode,
            mash_canonical_kmers=mash_canonical_kmers,
            show_progress=show_progress,
        )

        self._seq_store = seq_store
        self._limit = limit

        if parallel:
            if max_workers is None:
                max_workers = multiprocessing.cpu_count()
            self._numprocs = min(max_workers, multiprocessing.cpu_count())
        else:
            self._numprocs = 1

        self._calc_dist = (
            self._mash_dist if distance_mode == "mash" else self._euclidean_dist
        )

    def main(self, seq_names: list[str]) -> PhyloNode:
        """Construct a cluster tree for a collection of sequences.

        Parameters
        ----------
        seqs
            Sequence collection to form cluster tree for.

        Returns
        -------
        PhyloNode
            a cluster tree.
        """
        dstore = dvs_data_store.HDF5DataStore(self._seq_store, mode="r")

        self._executor = (
            get_reusable_executor(max_workers=self._numprocs)
            if self._numprocs != 1
            else nullcontext()
        )

        seq_arrays = {
            m.unique_id: m  # pylint: disable=not-callable
            for m in dstore.completed
            if m.unique_id in seq_names
        }
        seq_arrays = [seq_arrays[name] for name in seq_names]
        seq_arrays = seq_arrays[: self._limit] if self._limit else seq_arrays
        seq_arrays = [
            SeqArray(
                member.unique_id,
                member.read(),
                self._moltype,
                member.data_store.source,
            )
            for member in seq_arrays
        ]

        with self._progress, self._executor:
            distances = self._calc_dist(seq_arrays)
            return make_cluster_tree(seq_names, distances, progress=self._progress)


def compute_mash_chunk_distances(
    start_idx: int,
    stride: int,
    sketches: list[BottomSketch],
    k: int,
    sketch_size: int,
) -> tuple[int, dok_matrix]:
    """Find a subset of pairwise distances between sketches.

    Finds pairwise distances for the row with the given start index
    and each stride after that.

    Only finds lower triangular portion.

    Parameters
    ----------
    start_idx
        Start index for distance calculation.
    stride
        Index increment.
    sketches
        Sketches for pairwise distances.
    k
        kmer size.
    sketch_size
        Size of the sketches.

    Returns
    -------
    tuple[int, dok_matrix]
        The start index, and the calculated pairwise distances.
    """
    distances_chunk = dok_matrix((len(sketches), len(sketches)))
    for i in range(start_idx, len(sketches), stride):
        for j in range(i):
            distance = mash_distance(sketches[i], sketches[j], k, sketch_size)
            distances_chunk[i, j] = distance
    return start_idx, distances_chunk


def compute_euclidean_chunk_distances(
    start_idx: int,
    stride: int,
    kmer_seqs: list[KmerSeq],
) -> tuple[int, dok_matrix]:
    """Find a subset of pairwise distances between sketches.

    Finds pairwise distances for the row with the given start index
    and each stride after that.

    Only finds lower triangular portion.

    Parameters
    ----------
    start_idx
        Start index for distance calculation.
    stride
        Index increment.
    kmer_seqs
        kmer sequences.

    Returns
    -------
    tuple[int, dok_matrix]
        The start index, and the calculated pairwise distances.
    """
    distances_chunk = dok_matrix((len(kmer_seqs), len(kmer_seqs)))
    for i in range(start_idx, len(kmer_seqs), stride):
        freq_i = numpy.array(kmer_seqs[i].kfreqs)
        for j in range(i):
            freq_j = numpy.array(kmer_seqs[j].kfreqs)
            distance = euclidean_distance(freq_i, freq_j)
            distances_chunk[i, j] = distance
    return start_idx, distances_chunk



# pylint: disable=not-callable
import pytest
from cogent3 import make_tree
from cogent3.core.new_alignment import SequenceCollection

from diverse_seq.cluster import dvs_ctree, dvs_par_ctree


def check_ctree_app(app: dvs_ctree | dvs_par_ctree, seqs: SequenceCollection) -> None:
    tree = app(seqs.take_seqs(["Human", "Chimpanzee", "Rhesus", "Horse"]))
    expected = make_tree("(((Human, Chimpanzee), Rhesus), Horse);")
    assert tree.same_topology(expected)

    tree = app(seqs.take_seqs(["Human", "Chimpanzee", "Manatee", "Dugong"]))
    expected = make_tree("((Human, Chimpanzee), (Manatee, Dugong));")
    assert tree.same_topology(expected)

    tree = app(seqs.take_seqs(["Human", "Chimpanzee", "Manatee", "Dugong", "Rhesus"]))
    expected = make_tree("(((Human, Chimpanzee), Rhesus), (Manatee, Dugong));")
    assert tree.same_topology(expected)


@pytest.mark.parametrize("sketch_size", [400, 4e50])
def test_ctree_mash(unaligned_seqs: SequenceCollection, sketch_size: int) -> None:
    app = dvs_ctree(k=16, sketch_size=sketch_size, distance_mode="mash")
    check_ctree_app(app, unaligned_seqs)


def test_ctree_euclidean(unaligned_seqs: SequenceCollection) -> None:
    app = dvs_ctree(k=5, distance_mode="euclidean")
    check_ctree_app(app, unaligned_seqs)


@pytest.mark.parametrize("max_workers", [1, 4])
def test_ctree_mash_parallel(
    unaligned_seqs: SequenceCollection,
    max_workers: int,
) -> None:
    app = dvs_par_ctree(
        k=16,
        sketch_size=400,
        max_workers=max_workers,
        parallel=True,
        distance_mode="mash",
    )
    check_ctree_app(app, unaligned_seqs)


@pytest.mark.parametrize("max_workers", [1, 4])
def test_ctree_euclidean_parallel(
    unaligned_seqs: SequenceCollection,
    max_workers: int,
) -> None:
    app = dvs_par_ctree(
        k=5,
        max_workers=max_workers,
        parallel=True,
        distance_mode="euclidean",
    )
    check_ctree_app(app, unaligned_seqs)



import functools
import inspect
import io
import pathlib
from collections.abc import Sequence

import h5py
import hdf5plugin
import numpy
import numpy.typing
from cogent3.app.data_store import (
    DataMember,
    DataStoreABC,
    Mode,
    StrOrBytes,
)
from scitrack import get_text_hexdigest

_NOT_COMPLETED_TABLE = "not_completed"
_LOG_TABLE = "logs"
_MD5_TABLE = "md5"
_H5_STRING_DTYPE = h5py.string_dtype()


_HDF5_BLOSC2_KWARGS = hdf5plugin.Blosc2(
    cname="blosclz",
    clevel=3,
    filters=hdf5plugin.Blosc2.BITSHUFFLE,
)


class HF5FileWrapper:
    """virtualizes a HDF5 file so that it can be either on disk or in memory
    behaving consistently as a context manager"""

    def __init__(
        self,
        *,
        source: str | pathlib.Path,
        mode: Mode = "r",
        in_memory: bool = False,
    ):
        self.in_memory = in_memory
        mode = "w" if in_memory else mode
        self.mode = Mode(mode)
        self.source = pathlib.Path(source)
        if self.mode == Mode.r and not self.source.exists():
            raise OSError(f"{self.source!s} not found")
        if in_memory:
            mode = "w"
            in_memory = io.BytesIO()

        self._file = None
        self._init_file(in_memory)

    @functools.singledispatchmethod
    def _init_file(self, in_memory: io.BytesIO) -> None:
        self._file = h5py.File(in_memory, mode=self.mode.name)
        self._file.create_group(_NOT_COMPLETED_TABLE)
        self._file.create_group(_LOG_TABLE)
        self._file.create_group(_MD5_TABLE)
        self._file.attrs["source"] = str(self.source)

    @_init_file.register
    def _(self, in_memory: bool) -> None:
        if self.mode == Mode.w:
            with h5py.File(self.source, mode="w") as f:
                f.create_group(_NOT_COMPLETED_TABLE)
                f.create_group(_LOG_TABLE)
                f.create_group(_MD5_TABLE)
                f.attrs["source"] = str(self.source)
            self.mode = Mode.a

    def __enter__(self):
        if not self.in_memory:
            self._file = h5py.File(self.source, mode=self.mode.name)
        return self._file

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.in_memory:
            self._file.close()
        return False

    def close(self):
        if self.in_memory:
            self._file.close()


class HDF5DataStore(DataStoreABC):
    """Stores array data in HDF5 data sets. Associated information is
    stored as attributed on data sets.
    """

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        init_sig = inspect.signature(cls.__init__)
        bargs = init_sig.bind_partial(cls, *args, **kwargs)
        bargs.apply_defaults()
        init_vals = bargs.arguments
        init_vals.pop("self", None)
        obj._init_vals = init_vals
        return obj

    def __init__(
        self,
        source: str | pathlib.Path,
        mode: Mode = "r",
        limit: int = None,
        in_memory: bool = False,
    ) -> None:
        self._hf5_file = HF5FileWrapper(source=source, mode=mode, in_memory=in_memory)
        self._source = pathlib.Path(source)
        self._mode = self._hf5_file.mode
        self._limit = limit
        self._completed = []

    def __getstate__(self):
        init_vals = self._init_vals.copy()
        init_vals["mode"] = "w" if init_vals["in_memory"] else "r"
        return init_vals

    def __setstate__(self, state):
        obj = self.__class__(**state)
        self.__dict__.update(obj.__dict__)

    @property
    def source(self) -> pathlib.Path:
        """string that references connecting to data store"""
        return self._hf5_file.source

    @property
    def mode(self) -> Mode:
        """string that references datastore mode, override in subclass constructor"""
        return self._mode

    @property
    def limit(self):
        return self._limit

    def read(self, unique_id: str) -> numpy.ndarray:
        """reads and return array data corresponding to identifier"""
        with self._hf5_file as f:
            data = f[unique_id]
            out = numpy.empty(len(data), dtype=numpy.uint8)
            data.read_direct(out)
            return out

    def get_attrs(self, unique_id: str) -> dict:
        """return all data set attributes connected to an identifier"""
        with self._hf5_file as f:
            data = f[unique_id]

        return dict(data.attrs.items())

    def _write(
        self,
        *,
        subdir: str,
        unique_id: str,
        data: numpy.ndarray,
        **kwargs,
    ) -> DataMember:
        if subdir == _LOG_TABLE:
            return None

        with self._hf5_file as f:
            path = f"{subdir}/{unique_id}" if subdir else unique_id
            dset = f.create_dataset(path, data=data, dtype="u1", **_HDF5_BLOSC2_KWARGS)

            if subdir == _NOT_COMPLETED_TABLE:
                member = DataMember(
                    data_store=self,
                    unique_id=pathlib.Path(_NOT_COMPLETED_TABLE) / unique_id,
                )

            elif not subdir:
                member = DataMember(data_store=self, unique_id=unique_id)
                for key, value in kwargs.items():
                    dset.attrs[key] = value

            md5 = get_text_hexdigest(data.tobytes())
            f.create_dataset(
                f"{_MD5_TABLE}/{unique_id}",
                data=md5,
                dtype=_H5_STRING_DTYPE,
            )
            del data

        return member

    def write(self, *, unique_id: str, data: numpy.ndarray, **kwargs) -> DataMember:
        """Writes a completed record to a dataset in the HDF5 file.

        Parameters
        ----------
        unique_id
            The unique identifier for the record. This will be used
            as the name of the dataset in the HDF5 file.
        data
            The data to be stored in the record.
        kwargs
            Any additional keyword arguments will be stored as attributes on the
            dataset. Each key-value pair in kwargs corresponds to an attribute
            name and its value.

        Returns
        -------
        DataMember
            A DataMember object representing the new record.

        Notes
        -----
        Drops any not-completed member corresponding to this identifier
        """
        member = self._write(subdir="", unique_id=unique_id, data=data, **kwargs)
        self.drop_not_completed(unique_id=unique_id)
        if member is not None:
            self._completed.append(member)
        return member

    def write_not_completed(self, *, unique_id: str, data: StrOrBytes) -> None: ...

    def write_log(self, *, unique_id: str, data: StrOrBytes) -> None: ...

    def drop_not_completed(self, *, unique_id: str | None = None) -> None: ...

    def md5(self, unique_id: str) -> str | None:
        with self._hf5_file as f:
            if f"{_MD5_TABLE}/{unique_id}" in f:
                dset = f[f"{_MD5_TABLE}/{unique_id}"]
                return dset[()].decode("utf-8")
        return None

    @property
    def completed(self) -> list[DataMember]:
        if not self._completed:
            r = []
            with self._hf5_file as f:
                for name in f.keys():
                    if name in (_LOG_TABLE, _NOT_COMPLETED_TABLE, _MD5_TABLE):
                        continue

                    m = DataMember(data_store=self, unique_id=name)
                    r.append(m)
            self._completed = r
        return self._completed

    @property
    def logs(self) -> list[DataMember]:
        return []

    @property
    def not_completed(self) -> list[DataMember]:
        return []

    def __del__(self):
        self.close()

    def close(self):
        """closes the hdf5 file"""
        # hdf5 dumps content to stdout if resource already closed, so
        # we trap that here, and capture expected exceptions raised in the
        # process


def get_seqids_from_store(
    seq_store: str | pathlib.Path,
) -> list[str]:
    """return the list of seqids in a sequence store"""
    dstore = HDF5DataStore(seq_store, mode="r")
    return [m.unique_id for m in dstore.completed]


def get_ordered_records(
    seq_store: HDF5DataStore,
    seq_names: Sequence[str],
) -> list[DataMember]:
    """Returns ordered data store records given a seqeunce
    of sequence names.

    Parameters
    ----------
    seq_store
        The data store to load from.
    seq_names
        The ordered sequence names.

    Returns
    -------
    list[DataMember]
        Ordered data member records.
    """
    records = {m.unique_id: m for m in seq_store.completed if m.unique_id in seq_names}
    return [records[name] for name in seq_names]



"""defines basic data type for storing an individual sequence record"""

import dataclasses
import functools
from collections.abc import Iterator
from math import fabs, isclose

import numba
import typing_extensions
from attrs import asdict, define, field, validators
from cogent3 import get_moltype
from cogent3.app import composable
from cogent3.app import data_store as c3_data_store
from cogent3.app import typing as c3_types
from cogent3.core import new_sequence as c3_new_seq
from cogent3.core import sequence as c3_seq
from numpy import (
    array,
    dtype,
    errstate,
    log2,
    min_scalar_type,
    nan_to_num,
    ndarray,
    uint8,
    uint64,
    zeros,
)
from numpy import divmod as np_divmod

from diverse_seq import util as dvs_utils

NumType = float | int
PosDictType = dict[int, NumType]


@functools.singledispatch
def _gettype(name) -> type:
    try:
        return name.type
    except AttributeError as e:
        msg = f"type {type(name)} not supported"
        raise TypeError(msg) from e


@_gettype.register
def _(name: str) -> type:
    import numpy

    if name[-1].isdigit():
        return getattr(numpy, name)
    return {"int": int, "float": float}[name]


@_gettype.register
def _(name: type) -> type:
    return name


@dataclasses.dataclass(slots=True)
class lazy_kmers:
    data: c3_data_store.DataMember | ndarray
    k: int
    dtype: type = uint64
    num_states: int = dataclasses.field(init=False)
    moltype: dataclasses.InitVar[str] = dataclasses.field(default="dna")

    def __post_init__(self, moltype: str) -> None:
        self.num_states = len(_get_canonical_states(moltype))

    def __array__(
        self,
        dtype: dtype | None = None,
        copy: bool | None = None,
    ) -> ndarray[int]:
        data = self.data if isinstance(self.data, ndarray) else self.data.read()
        return kmer_counts(data, self.num_states, self.k, dtype=self.dtype)


@functools.singledispatch
def _make_data(data, size: int | None = None, dtype: type = None) -> ndarray:
    raise NotImplementedError


@_make_data.register
def _(data: ndarray, size: int | None = None, dtype: type = None) -> ndarray:
    return data


@_make_data.register
def _(data: lazy_kmers, size: int | None = None, dtype: type = None) -> ndarray:
    return data


@_make_data.register
def _(data: None, size: int | None = None, dtype: type = None) -> ndarray:
    return zeros(size, dtype=dtype)


@_make_data.register
def _(data: dict, size: int | None = None, dtype: type = None) -> ndarray:
    result = zeros(size, dtype=dtype)
    for i, n in data.items():
        if isclose(n, 0):
            continue
        result[i] = n
    return result


# convert this to a dataclass
# write a wrapper for the data store member so it has a __array__ method
@define(slots=True)
class vector:
    data: ndarray
    vector_length: int
    default: NumType | None = field(init=False)
    dtype: type = float
    source: str = None
    name: str = None

    def __init__(
        self,
        *,
        vector_length: int,
        data: ndarray = None,
        dtype: type = float,
        source: str = None,
        name: str = None,
    ) -> None:
        """
        Parameters
        ----------
        vector_length
            num_states**k
        data
            dict of {k-mer index: NumType}
        dtype
        """
        self.vector_length = vector_length
        dtype = _gettype(dtype)
        self.dtype = dtype

        data = _make_data(data, size=vector_length, dtype=dtype)
        self.default = dtype(0)
        self.data = data
        self.source = source
        self.name = name

    def __setitem__(self, index: int, value: NumType) -> None:
        self.data[index] = value

    def __getitem__(self, index: int) -> NumType:
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[NumType]:
        yield from self.data

    def __getstate__(self) -> dict:
        return asdict(self)

    def __setstate__(self, data: dict) -> typing_extensions.Self:
        for k, v in data.items():
            setattr(self, k, v)
        return self

    def __sub__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        data = self.data - other
        return self.__class__(
            data=data,
            vector_length=self.vector_length,
            dtype=self.dtype,
        )

    def __isub__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        self.data -= other
        return self

    def __add__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        # we are creating a new instance
        data = self.data + other
        return self.__class__(
            data=data,
            vector_length=self.vector_length,
            dtype=self.dtype,
        )

    def __iadd__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        self.data += other
        return self

    def __truediv__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        # we are creating a new instance
        with errstate(divide="ignore", invalid="ignore"):
            data = nan_to_num(self.data / other, nan=0.0, copy=False)
        return self.__class__(data=data, vector_length=self.vector_length, dtype=float)

    def __itruediv__(self, other: typing_extensions.Self) -> typing_extensions.Self:
        with errstate(divide="ignore", invalid="ignore"):
            data = nan_to_num(self.data / other, nan=0.0, copy=False)
        self.dtype = float
        self.data = data
        return self

    def sum(self) -> NumType:
        return self.data.sum()

    def iter_nonzero(self) -> Iterator[NumType]:
        yield from (v for v in self.data if v)

    @property
    def entropy(self) -> float:
        non_zero = self.data[self.data > 0]
        kfreqs = non_zero if self.dtype == float else non_zero / non_zero.sum()
        # taking absolute value due to precision issues
        return fabs(-(kfreqs * log2(kfreqs)).sum())

    def __array__(
        self,
        dtype: dtype | None = None,
        copy: bool | None = None,
    ) -> ndarray[int]:
        if not isinstance(self.data, ndarray):
            self.data = array(self.data)
        return self.data


@numba.jit(nopython=True)
def coord_conversion_coeffs(num_states, k):  # pragma: no cover
    """coefficients for multi-dimensional coordinate conversion into 1D index"""
    return array([num_states ** (i - 1) for i in range(k, 0, -1)])


@numba.jit(nopython=True)
def coord_to_index(coord, coeffs):  # pragma: no cover
    """converts a multi-dimensional coordinate into a 1D index"""
    return (coord * coeffs).sum()


@numba.jit(nopython=True)
def index_to_coord(index, coeffs):  # pragma: no cover
    """converts a 1D index into a multi-dimensional coordinate"""
    ndim = len(coeffs)
    coord = zeros(ndim, dtype=uint64)
    remainder = index
    for i in range(ndim):
        n, remainder = np_divmod(remainder, coeffs[i])
        coord[i] = n
    return coord


def indices_to_seqs(indices: ndarray, states: bytes, k: int) -> list[str]:
    """convert indices from k-dim into sequence

        Parameters
    ----------
    indices
        array of (len(states), k) dimensioned indices
    states
        the ordered characters, e.g. b"TCAG"
    """
    arr = indices_to_bytes(indices, states, k)
    return [bytearray(kmer).decode("utf8") for kmer in arr]


@numba.jit(nopython=True)
def indices_to_bytes(
    indices: ndarray,
    states: bytes,
    k: int,
) -> ndarray:  # pragma: no cover
    """convert indices from k-dim into bytes

        Parameters
    ----------
    indices
        array of (len(states), k) dimensioned indices
    states
        the ordered characters, e.g. b"TCAG"
    k
        dimensions

    Raises
    -----
    IndexError if an index not in states

    Returns
    -------
    uint8 array.

    Notes
    -----
    Use indices_to_seqs to have the results returned as strings
    """
    result = zeros((len(indices), k), dtype=uint8)
    coeffs = coord_conversion_coeffs(len(states), k)
    num_states = len(states)
    for i in range(len(indices)):
        index = indices[i]
        coord = index_to_coord(index, coeffs)
        for j in range(k):
            if coord[j] >= num_states:
                msg = "index out of character range"
                raise IndexError(msg)
            result[i][j] = states[coord[j]]

    return result


@numba.jit(nopython=True)
def kmer_counts(
    seq: ndarray,
    num_states: int,
    k: int,
    dtype: dtype = uint64,
) -> ndarray:  # pragma: no cover
    """return freqs of valid k-mers using 1D indices

    Parameters
    ----------
    seq
        numpy array of uint, assumed that canonical characters have
        sequential indexes which are all < num_states
    result
        array to be written into
    num_states
        defines range of possible ints at a position
    k
        k-mer size

    Returns
    -------
    result
    """
    coeffs = coord_conversion_coeffs(num_states, k)
    kfreqs = zeros(num_states**k, dtype=dtype)
    skip_until = 0
    for i in range(k):
        if seq[i] >= num_states:
            skip_until = i + 1

    for i in range(len(seq) - k + 1):
        if seq[i + k - 1] >= num_states:
            skip_until = i + k

        if i < skip_until:
            continue

        kmer = seq[i : i + k]
        index = (kmer * coeffs).sum()
        kfreqs[index] += 1

    return kfreqs


def _gt_zero(instance, attribute, value):
    if value <= 0:
        msg = f"must be > 0, not {value}"
        raise ValueError(msg)


@functools.singledispatch
def _make_kcounts(data) -> vector:
    msg = f"type {type(data)} not supported"
    raise TypeError(msg)


@_make_kcounts.register
def _(data: ndarray):
    nonzero = {i: v for i, v in enumerate(data.tolist()) if v}
    return vector(
        vector_length=len(data),
        data=nonzero,
        dtype=_gettype(data.dtype.name),
    )


@_make_kcounts.register
def _(data: vector):
    return data


@dataclasses.dataclass(frozen=True)
class SeqArray:
    """A SeqArray stores an array of indices that map to the canonical characters
    of the moltype of the original sequence. Use rep_seq.util.arr2str() to
    recapitulate the original sequence."""

    seqid: str
    data: ndarray
    moltype: str
    source: str = None

    def __len__(self) -> int:
        return len(self.data)


@define(slots=True, order=True, hash=True)
class KmerSeq:
    """representation of a single sequence as kmer counts"""

    kcounts: vector = field(eq=False, converter=_make_kcounts)
    name: str = field(validator=validators.instance_of(str), eq=True)
    delta_jsd: float = field(
        init=False,
        validator=validators.instance_of(float),
        default=0.0,
        eq=False,
    )

    @property
    def size(self) -> int:
        return self.kcounts.vector_length

    @functools.cached_property
    def entropy(self) -> float:
        return self.kfreqs.entropy

    @functools.cached_property
    def kfreqs(self) -> vector:
        kcounts = array(self.kcounts)
        kcounts = kcounts.astype(float)
        kfreqs = kcounts / kcounts.sum()
        return vector(
            data=kfreqs,
            vector_length=len(kfreqs),
            dtype=float,
        )


@composable.define_app
class seq_to_seqarray:
    """Convert a cogent3 Sequence to a SeqArray"""

    def __init__(
        self,
        moltype: str = "dna",
    ) -> None:
        self.moltype = moltype
        self.str2arr = dvs_utils.str2arr(moltype=self.moltype)

    def main(self, seq: c3_types.SeqType) -> SeqArray:
        return SeqArray(
            seqid=seq.name,
            data=self.str2arr(str(seq)),
            moltype=self.moltype,
            source=seq.info.source or seq.name,
        )


@functools.singledispatch
def make_kmerseq(data, *, dtype: dtype, k: int, moltype: str) -> KmerSeq:
    msg = f"type {type(data)} not supported"
    raise TypeError(msg)


@make_kmerseq.register
def _(data: SeqArray, *, dtype: dtype, k: int, moltype: str) -> KmerSeq:
    vec = lazy_kmers(
        data=data.data,
        k=k,
        moltype=moltype,
        dtype=dtype,
    )
    kwargs = {
        "vector_length": vec.num_states,
        "dtype": dtype,
        "source": data.source,
        "name": data.seqid,
        "data": vec,
    }

    return KmerSeq(
        kcounts=vector(**kwargs),
        name=kwargs["name"],
    )


@make_kmerseq.register
def _(data: c3_data_store.DataMember, *, dtype: dtype, k: int, moltype: str) -> KmerSeq:
    vec = lazy_kmers(
        data=data,
        k=k,
        moltype=moltype,
        dtype=dtype,
    )
    kwargs = {
        "vector_length": vec.num_states,
        "dtype": dtype,
        "source": data.data_store.source,
        "name": data.unique_id,
        "data": vec,
    }

    return KmerSeq(
        kcounts=vector(**kwargs),
        name=kwargs["name"],
    )


@make_kmerseq.register
def _(data: c3_seq.Sequence, *, dtype: dtype, k: int, moltype: str) -> KmerSeq:
    cnvrt = dvs_utils.str2arr(moltype=moltype)
    vec = lazy_kmers(
        data=cnvrt(str(data)),  # pylint: disable=not-callable
        k=k,
        moltype=moltype,
        dtype=dtype,
    )
    kwargs = {
        "vector_length": vec.num_states,
        "dtype": dtype,
        "source": data.info.source,
        "name": data.name,
        "data": vec,
    }

    return KmerSeq(
        kcounts=vector(**kwargs),
        name=kwargs["name"],
    )


@make_kmerseq.register
def _(data: c3_new_seq.Sequence, *, dtype: dtype, k: int, moltype: str) -> KmerSeq:
    cnvrt = dvs_utils.str2arr(moltype=moltype)
    vec = lazy_kmers(
        data=cnvrt(str(data)),  # pylint: disable=not-callable
        k=k,
        moltype=moltype,
        dtype=dtype,
    )
    kwargs = {
        "vector_length": vec.num_states,
        "dtype": dtype,
        "source": data.info.source,
        "name": data.name,
        "data": vec,
    }

    return KmerSeq(
        kcounts=vector(**kwargs),
        name=kwargs["name"],
    )


class _make_kmerseq_init:
    def __init__(self, k: int, moltype: str) -> None:
        """
        Parameters
        ----------
        k
            size of k-mers
        moltype
            cogent3 molecular type
        """
        self.moltype = moltype
        self.k = k
        self.num_states = len(_get_canonical_states(moltype))
        self.dtype = min_scalar_type(self.num_states**k)


@composable.define_app
class seqarray_to_kmerseq(_make_kmerseq_init):
    """converts a sequence numpy array to a KmerSeq"""

    def main(self, seq: SeqArray) -> KmerSeq:
        return make_kmerseq(seq, dtype=self.dtype, k=self.k, moltype=self.moltype)


@composable.define_app
class member_to_kmerseq(_make_kmerseq_init):
    """creates a KmerSeq from a HDF5 datastore member"""

    def main(self, member: c3_data_store.DataMember) -> KmerSeq:
        return make_kmerseq(member, dtype=self.dtype, k=self.k, moltype=self.moltype)


@functools.cache
def _get_canonical_states(moltype: str) -> bytes:
    mt = get_moltype(moltype)
    canonical = "".join(mt.alphabet)
    v = mt.alphabet.to_indices(canonical)
    if not (0 <= min(v) < max(v) < len(canonical)):
        msg = f"indices of canonical states {canonical} not sequential {v}"
        raise ValueError(msg)
    return canonical.encode("utf8")



# pylint: disable=not-callable

from diverse_seq.distance import dvs_dist


def test_euclidean_distance(unaligned_seqs):
    app = dvs_dist(
        "euclidean",
        k=3,
    )

    unaligned_seqs = unaligned_seqs.take_seqs(
        ["Human", "Chimpanzee", "Manatee", "Dugong", "Rhesus"],
    )
    dists = app(unaligned_seqs)

    assert (
        dists["Human", "Chimpanzee"] < dists["Human", "Dugong"]
    )  # chimpanzee closer than rhesus
    assert (
        dists["Human", "Rhesus"] < dists["Human", "Manatee"]
    )  # rhesus closer than manatee
    assert (
        dists["Human", "Rhesus"] < dists["Human", "Dugong"]
    )  # rhesus closer than dugong

    assert (
        dists["Chimpanzee", "Rhesus"] < dists["Chimpanzee", "Manatee"]
    )  # rhesus closer than manatee
    assert (
        dists["Chimpanzee", "Rhesus"] < dists["Chimpanzee", "Dugong"]
    )  # rhesus closer than dugong

    assert (
        dists["Manatee", "Dugong"] < dists["Manatee", "Rhesus"]
    )  # dugong closer than rhesus


def test_mash_distance(unaligned_seqs):
    app = dvs_dist(
        "mash",
        k=16,
        sketch_size=400,
        mash_canonical_kmers=True,
    )

    unaligned_seqs = unaligned_seqs.take_seqs(
        ["Human", "Chimpanzee", "Manatee", "Dugong", "Rhesus"],
    )
    dists = app(unaligned_seqs)

    assert (
        dists["Human", "Chimpanzee"] < dists["Human", "Dugong"]
    )  # chimpanzee closer than rhesus
    assert (
        dists["Human", "Rhesus"] < dists["Human", "Manatee"]
    )  # rhesus closer than manatee
    assert (
        dists["Human", "Rhesus"] < dists["Human", "Dugong"]
    )  # rhesus closer than dugong

    assert (
        dists["Chimpanzee", "Rhesus"] < dists["Chimpanzee", "Manatee"]
    )  # rhesus closer than manatee
    assert (
        dists["Chimpanzee", "Rhesus"] < dists["Chimpanzee", "Dugong"]
    )  # rhesus closer than dugong

    assert (
        dists["Manatee", "Dugong"] < dists["Manatee", "Rhesus"]
    )  # dugong closer than rhesus



import sys

import numpy
import pytest
from cogent3 import get_moltype
from numpy.testing import assert_allclose

from diverse_seq import util as dvs_util


def test_str2arr():
    dna = get_moltype("dna")
    s = "ACGTT"
    expect = dna.alphabet.to_indices(s)
    app = dvs_util.str2arr()
    g = app(s)  # pylint: disable=not-callable
    assert (g == expect).all()
    g = app("ACGNT")  # pylint: disable=not-callable
    assert g[-2] > 3  # index for non-canonical character > num_states


@pytest.mark.parametrize("seq", ("ACGTT", "ACGNT", "AYGTT", ""))
def test_arr2str(seq):
    app = dvs_util.str2arr() + dvs_util.arr2str()
    got = app(seq)
    assert got == seq


@pytest.mark.parametrize(
    "suffix",
    ("fa", "fasta", "genbank", "gbk", "gbk.gz", "fna.bz2"),
)
def test_get_seq_format(suffix):
    assert dvs_util.get_seq_file_format(suffix) in ("fasta", "genbank")


@pytest.mark.parametrize("suffix", ("gbkgz", "paml"))
def test_get_seq_format_unknown(suffix):
    assert dvs_util.get_seq_file_format(suffix) is None


def test_determine_chunk_size():
    got = list(dvs_util.determine_chunk_size(10, 3))
    assert got == [4, 3, 3]


def test_chunked():
    data = list(range(10))
    got = list(dvs_util.chunked(data, 3))
    expect = [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9]]
    assert got == expect


def test_summary_stats():
    data = numpy.random.randint(low=0, high=5, size=100)
    stats = dvs_util.summary_stats(data)
    assert_allclose(stats.mean, data.mean())
    assert_allclose(stats.std, data.std(ddof=1))
    assert_allclose(stats.cov, data.std(ddof=1) / data.mean())


@pytest.fixture(params=(False, True))
def includes(request, tmp_path):
    names = ["a", "b", "c"]
    if request.param:
        inc = tmp_path / "includes.txt"
        inc.write_text("\n".join(names), encoding="utf-8")
        return inc
    return ",".join(names)


def test_parse_include_arg(includes):
    got = dvs_util._comma_sep_or_file(None, None, includes)
    assert got == ["a", "b", "c"]


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Not sure if this test will work on Windows",
)
def test_print_colour(capsys):
    dvs_util.print_colour("hello", "red")
    got = capsys.readouterr().out
    assert "hello" in got


def test_export_demo(tmp_dir):
    got = dvs_util.get_sample_data_path()
    assert got.exists()
    assert got.name == "brca1.fa"



import pathlib
import sys

import h5py
import pytest
from click.testing import CliRunner
from cogent3 import load_table, load_tree, load_unaligned_seqs

from diverse_seq.cli import ctree as dvs_ctree
from diverse_seq.cli import max as dvs_max
from diverse_seq.cli import nmost as dvs_nmost
from diverse_seq.cli import prep as dvs_prep
from diverse_seq.data_store import (
    _LOG_TABLE,
    _MD5_TABLE,
    _NOT_COMPLETED_TABLE,
    HDF5DataStore,
)

__author__ = "Gavin Huttley"
__credits__ = ["Gavin Huttley"]

DATADIR = pathlib.Path(__file__).parent / "data"


@pytest.fixture(scope="function")
def tmp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("dvgt")


@pytest.fixture(scope="function")
def tmp_path(tmp_path_factory):
    return tmp_path_factory.mktemp("seqdata")


@pytest.fixture(scope="session")
def runner():
    """exportrc works correctly."""
    return CliRunner()


@pytest.fixture(scope="function")
def rna_seq_path(tmp_dir, seq_path):
    seqs = load_unaligned_seqs(seq_path, moltype="rna")
    path = tmp_dir / "test_rna.fasta"
    seqs.write(path)
    return path


@pytest.fixture(scope="function")
def seq_dir(tmp_path, seq_path):
    seqs = load_unaligned_seqs(seq_path, moltype="dna")
    for s in seqs.iter_seqs():
        fn = tmp_path / f"{s.name}.fasta"
        fn.write_text(s.to_fasta(), encoding="utf-8")
    return tmp_path


@pytest.fixture(scope="session")
def processed_seq_path():
    return DATADIR / "brca1.dvseqs"


def _checked_output(path, eval_rows=None, eval_header=None):
    table = load_table(str(path))
    assert table.shape[1] == 2, table.shape[1]
    if eval_rows:
        assert eval_rows(table.shape[0])

    if eval_header:
        assert eval_header(table.header)


def _checked_h5_dstore(path, source=None):
    # TODO: how to more thoroughly interrogate output
    dstore = HDF5DataStore(path)
    assert len(dstore.completed) > 0
    assert len(dstore.not_completed) == 0


def test_defaults(runner, tmp_dir, processed_seq_path):
    outpath = tmp_dir / "test_defaults.tsv"
    args = f"-s {processed_seq_path} -o {outpath}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(str(outpath))


@pytest.mark.parametrize("min_size", (2, 5))
def test_min_size(runner, tmp_dir, processed_seq_path, min_size):
    outpath = tmp_dir / "test_min_size.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -z {min_size}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath, eval_rows=lambda x: x >= min_size)


@pytest.mark.parametrize("max_size", (5, 7))
def test_max_size(runner, tmp_dir, processed_seq_path, max_size):
    outpath = tmp_dir / "test_max_size.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -z 3 -zp {max_size}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath, eval_rows=lambda x: x <= max_size)


def test_max_include(runner, tmp_dir, processed_seq_path):
    outpath = tmp_dir / "test_include.tsv"
    args = (
        f"-s {processed_seq_path} -o {outpath} -z 3 -zp 5 -L 10 -k 1 -i Human".split()
    )
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    got = load_table(outpath)
    assert "Human" in got.columns["names"]


@pytest.mark.parametrize("size", (5, 7))
def test_min_eq_max(runner, tmp_dir, processed_seq_path, size):
    outpath = tmp_dir / "test_min_eq_max.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -z {size} -zp {size}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath, eval_rows=lambda x: x == size)


def test_min_gt_max_fail(runner, tmp_dir, processed_seq_path):
    outpath = tmp_dir / "test_min_gt_max_fail.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -zp 2".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code != 0, r.output


@pytest.mark.parametrize("stat", ("stdev", "cov"))
def test_stat(runner, tmp_dir, processed_seq_path, stat):
    outpath = tmp_dir / "test_defaults.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -st {stat}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath)


@pytest.mark.parametrize("k", (1, 3, 7))
def test_k(runner, tmp_dir, processed_seq_path, k):
    outpath = tmp_dir / "test_defaults.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -k {k}".split()
    r = runner.invoke(dvs_max, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath)


def test_nmost(runner, tmp_dir, processed_seq_path):
    outpath = tmp_dir / "test_defaults.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -k 1 -n 5".split()
    r = runner.invoke(dvs_nmost, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_output(outpath, eval_rows=lambda x: x == 5)


def test_nmost_include(runner, tmp_dir, processed_seq_path):
    outpath = tmp_dir / "test_include.tsv"
    args = f"-s {processed_seq_path} -o {outpath} -k 1 -n 5 -L 10 -i Human".split()
    r = runner.invoke(dvs_nmost, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    got = load_table(outpath)
    assert "Human" in got.columns["names"]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Test skipped on Windows")
def test_prep_seq_file(runner, tmp_dir, seq_path):
    outpath = tmp_dir / "test_prep_seq_file.dvseqs"
    args = f"-s {seq_path} -o {outpath} -sf fasta".split()
    r = runner.invoke(dvs_prep, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_h5_dstore(str(outpath))


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Test skipped on Windows")
def test_prep_outpath_without_suffix(runner, tmp_dir, seq_path):
    outpath = tmp_dir / "test_prep_outpath_without_suffix"
    args = f"-s {seq_path} -o {outpath}".split()
    r = runner.invoke(dvs_prep, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output
    _checked_h5_dstore(str(outpath) + ".dvseqs")


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Test skipped on Windows")
def test_prep_force_override(runner, tmp_dir, seq_path):
    outpath = tmp_dir / "test_prep_force_override.dvseqs"
    args = f"-s {seq_path} -o {outpath}".split()

    # Run prep once, it should succeed
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    _checked_h5_dstore(str(outpath))

    # with the force write flag, it should succeed
    args += ["-F"]
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    _checked_h5_dstore(str(outpath))


def test_prep_max_rna(runner, tmp_dir, rna_seq_path):
    outpath = tmp_dir / "test_prep_max_rna.dvseqs"
    args = f"-s {rna_seq_path} -o {outpath} -m rna".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    _checked_h5_dstore(str(outpath))

    max_outpath = tmp_dir / "test_prep_rna.tsv"
    max_args = f"-s {outpath} -o {max_outpath}".split()
    r = runner.invoke(dvs_max, max_args)
    assert r.exit_code == 0, r.output
    _checked_output(str(max_outpath))


def test_prep_source_from_file(runner, tmp_dir, seq_path):
    outpath = tmp_dir / "test_prep_source_from_file.dvseqs"
    args = f"-s {seq_path} -o {outpath} -sf fasta".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0


def test_prep_source_from_directory(runner, tmp_dir, seq_dir):
    outpath = tmp_dir / "test_prep_source_from_directory.dvseqs"
    args = f"-s {seq_dir} -o {outpath} -sf fasta".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    with h5py.File(outpath, mode="r") as f:
        sources = {
            dset.attrs["source"]
            for name, dset in f.items()
            if name not in (_LOG_TABLE, _MD5_TABLE, _NOT_COMPLETED_TABLE)
        }

    assert sources == {str(seq_dir)}


@pytest.mark.parametrize(
    ("distance", "k", "sketch_size"),
    [("mash", 16, 400), ("euclidean", 5, None)],
)
@pytest.mark.parametrize("max_workers", [1, 4])
def test_ctree(
    runner,
    tmp_dir,
    processed_seq_path,
    distance,
    k,
    sketch_size,
    max_workers,
):
    outpath = tmp_dir / "out.tre"

    args = f"-s {processed_seq_path} -o {outpath} -d {distance} -k {k} -np {max_workers} -hp"
    if sketch_size is not None:
        args += f" --sketch-size {sketch_size}"
    args = args.split()

    r = runner.invoke(dvs_ctree, args, catch_exceptions=False)
    assert r.exit_code == 0, r.output

    tree = load_tree(outpath)
    expected_tips = 55
    assert len(tree.tips()) == expected_tips


def test_prep_force(runner, tmp_path):
    import diverse_seq

    data = diverse_seq.load_sample_data()
    one = data.take_seqs(data.names[:20])
    two = data.take_seqs(data.names[20:])
    one_path = tmp_path / "one.fasta"
    two_path = tmp_path / "two.fasta"
    one.write(one_path)
    two.write(two_path)
    outpath = tmp_path / "test_prep_force.dvseqs"
    args = f"-s {one_path} -o {outpath}".split()

    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    # we try again but the file already exists, so
    # exit code should be 1
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 1, "did not fail when file already exists"

    args = f"-s {one_path} -o {outpath} -F".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output


@pytest.fixture(params=[True, False])
def prep_too_few(request, tmp_path):
    import diverse_seq

    data_path = tmp_path / "sample.fasta"
    return_dir = request.param
    # when processing a directory of files, prep
    # extracts a single seq from each file
    data = diverse_seq.load_sample_data()
    if not return_dir:
        ret_path = data_path
        data = data.take_seqs(data.names[:3])
    else:
        ret_path = data_path.parent

    data.write(data_path)
    return ret_path


def test_prep_too_few_seqs(prep_too_few, runner, tmp_path):
    out_path = tmp_path / "too_few.dvseqs"
    args = f"-s {prep_too_few} -sf fasta -o {out_path}".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 1, r.output


@pytest.fixture
def too_few_seqs(tmp_path):
    import diverse_seq

    data = diverse_seq.load_sample_data()
    data = data.take_seqs(data.names[:7])
    data_path = tmp_path / "sample.fasta"
    data.write(data_path)
    return data_path


def test_nmost_too_few_seqs(runner, too_few_seqs):
    outpath = too_few_seqs.parent / "too_few.dvseqs"
    args = f"-s {too_few_seqs} -o {outpath}".split()
    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    args = f"-n 8 -s {outpath} -o demo.tsv".split()
    # there are too few sequences
    # either as the minimum number for analysis, or below user set n
    # exit code should be 1
    r = runner.invoke(dvs_nmost, args)
    assert r.exit_code == 1, r.output


def test_max_too_few_seqs(runner, too_few_seqs):
    outpath = too_few_seqs.parent / "too_few.dvseqs"
    args = f"-s {too_few_seqs} -o {outpath}".split()

    r = runner.invoke(dvs_prep, args)
    assert r.exit_code == 0, r.output
    args = f"--min_size 8 -s {outpath} -o demo.tsv".split()
    # there are too few sequences
    # either as the minimum number for analysis, or below user set n
    # exit code should be 1
    r = runner.invoke(dvs_max, args)
    assert r.exit_code == 1, r.output



import heapq
import math
from collections.abc import Sequence
from typing import Literal, TypeAlias

import cogent3.app.typing as c3_types
import numba
import numpy as np
from cogent3.app.composable import define_app
from cogent3.evolve.fast_distance import DistanceMatrix
from rich.progress import Progress

from diverse_seq.record import (
    SeqArray,
    _get_canonical_states,
    make_kmerseq,
    seq_to_seqarray,
)

BottomSketch: TypeAlias = list[int]


@define_app
class dvs_dist:
    """Calculate pairwise kmer-based distances between sequences.
    Supported distances include mash distance, and euclidean distance
    based on kmer frequencies.
    """

    def __init__(
        self,
        distance_mode: Literal["mash", "euclidean"] = "mash",
        *,
        k: int = 12,
        sketch_size: int | None = 3000,
        moltype: str = "dna",
        mash_canonical_kmers: bool | None = None,
        show_progress: bool = False,
    ) -> None:
        """Initialise parameters for kmer distance calculation.

        Parameters
        ----------
        distance_mode
            mash distance or euclidean distance between kmer freqs
        k
            kmer size
        sketch_size
            size of sketches, by default None
        moltype
            moltype
        mash_canonical_kmers
            whether to use mash canonical kmers for mash distance
        show_progress
            whether to show progress bars

        Notes
        -----
        If mash_canonical_kmers is enabled when using the mash distance,
        kmers are considered identical to their reverse complement.

        References
        ----------
        .. [1] Ondov, B. D., Treangen, T. J., Melsted, P., Mallonee, A. B.,
           Bergman, N. H., Koren, S., & Phillippy, A. M. (2016).
           Mash: fast genome and metagenome distance estimation using MinHash.
           Genome biology, 17, 1-14.
        """
        if mash_canonical_kmers is None:
            mash_canonical_kmers = False

        if distance_mode not in ("mash", "euclidean"):
            msg = f"Unexpected distance {distance_mode!r}."
            raise ValueError(msg)

        if moltype not in ("dna", "rna") and mash_canonical_kmers:
            msg = "Canonical kmers only supported for dna sequences."
            raise ValueError(msg)

        if distance_mode == "mash" and sketch_size is None:
            msg = "Expected sketch size for mash distance measure."
            raise ValueError(msg)

        self._moltype = moltype
        self._k = k
        self._show_progress = show_progress

        self._s2a = seq_to_seqarray(moltype=moltype)
        if distance_mode == "mash":
            self._func = mash_distances
            kwargs = dict(
                k=self._k,
                sketch_size=sketch_size,
                num_states=len(_get_canonical_states(self._moltype)),
                mash_canonical=mash_canonical_kmers,
            )
        else:
            self._func = euclidean_distances
            kwargs = dict(k=self._k, moltype=self._moltype)
        self._func_kwargs = kwargs

    def main(
        self,
        seqs: c3_types.SeqsCollectionType,
    ) -> c3_types.PairwiseDistanceType:
        seqs = seqs.to_moltype(self._moltype)
        seq_arrays = [self._s2a(seqs.get_seq(name)) for name in seqs.names]  # pylint: disable=not-callable

        with Progress(disable=not self._show_progress) as progress:
            distances = self._func(
                seq_arrays,
                progress=progress,
                **self._func_kwargs,
            )

        return dists_to_distmatrix(distances, seqs.names)


def dists_to_distmatrix(
    distances: np.ndarray,
    names: Sequence[str],
) -> c3_types.PairwiseDistanceType:
    dist_dict = {}
    for i in range(1, len(distances)):
        for j in range(i):
            dist_dict[(names[i], names[j])] = distances[i, j]
    return DistanceMatrix(dist_dict)


def mash_distances(
    seq_arrays: list[SeqArray],
    k: int,
    sketch_size: int,
    num_states: int,
    *,
    mash_canonical: bool = False,
    progress: Progress | None = None,
) -> np.ndarray:
    """Calculates pairwise mash distances between sequences.

    Parameters
    ----------
    seq_arrays
        Sequence arrays.
    k
        kmer size.
    sketch_size
        sketch size.
    num_states
        number of states for each position.
    mash_canonical
        whether to use mash canonical representation of kmers,
        by default False
    progress
        progress bar, by default None

    Returns
    -------
    numpy.ndarray
        Pairwise mash distances between sequences.
    """
    if progress is None:
        progress = Progress(disable=True)

    seqs = [seq_array.data for seq_array in seq_arrays]

    sketches = mash_sketches(
        seqs,
        k,
        sketch_size,
        num_states,
        mash_canonical=mash_canonical,
        progress=progress,
    )

    distance_task = progress.add_task(
        "[green]Computing Pairwise Distances",
        total=len(sketches) * (len(sketches) - 1) // 2,
    )

    distances = np.zeros((len(sketches), len(sketches)))

    for i in range(1, len(sketches)):
        for j in range(i):
            distance = mash_distance(
                sketches[i],
                sketches[j],
                k,
                sketch_size,
            )
            distances[i, j] = distance
            distances[j, i] = distance

            progress.update(distance_task, advance=1)

    return distances


def mash_sketches(
    seq_arrays: Sequence[np.ndarray],
    k: int,
    sketch_size: int,
    num_states: int,
    *,
    mash_canonical: bool = False,
    progress: Progress | None = None,
) -> list[BottomSketch]:
    """Create sketch representations for a collection of sequence sequence arrays.

    Parameters
    ----------
    seq_arrays
        Sequence arrays.
    k
        kmer size.
    sketch_size
        sketch size.
    num_states
        number of states.
    mash_canonical
        whether to use mash canonical kmer representation, by default False
    progress
        progress bar, by default None
    Returns
    -------
    list[BottomSketch]
        Sketches for each sequence.
    """
    if progress is None:
        progress = Progress(disable=True)

    sketch_task = progress.add_task(
        "[green]Generating Sketches",
        total=len(seq_arrays),
    )

    bottom_sketches = [None for _ in range(len(seq_arrays))]

    # Compute sketches in serial
    for i, seq_array in enumerate(seq_arrays):
        bottom_sketches[i] = mash_sketch(
            seq_array,
            k,
            sketch_size,
            num_states,
            mash_canonical,
        )

        progress.update(sketch_task, advance=1)

    return bottom_sketches


@numba.njit
def mash_sketch(
    seq_array: np.ndarray,
    k: int,
    sketch_size: int,
    num_states: int,
    mash_canonical: bool,
) -> BottomSketch:  # pragma: no cover
    """Find the mash sketch for a sequence array.

    Parameters
    ----------
    seq_array
        The sequence array to find the sketch for.
    k
        kmer size.
    sketch_size
        Size of the sketch.
    num_states
        Number of possible states (e.g. GCAT gives 4 for DNA).
    mash_canonical
        Whether to use the mash canonical representation of kmers.

    Returns
    -------
    BottomSketch
        The bottom sketch for the given sequence seq_array.
    """
    kmer_hashes = np.unique(get_kmer_hashes(seq_array, k, num_states, mash_canonical))

    heap = [0]
    heap.pop()

    for kmer_hash in kmer_hashes:
        if len(heap) < sketch_size:
            heapq.heappush(heap, -kmer_hash)
        else:
            heapq.heappushpop(heap, -kmer_hash)
    kmer_hashes = [-kmer_hash for kmer_hash in heap]
    kmer_hashes.sort()
    return kmer_hashes


@numba.njit
def get_kmer_hashes(
    seq: np.ndarray,
    k: int,
    num_states: int,
    mash_canonical: bool,
) -> np.ndarray:  # pragma: no cover
    """Get the kmer hashes comprising a sequence.

    Parameters
    ----------
    seq
        A sequence.
    k
        kmer size.
    num_states
        Number of states allowed for sequence type.
    mash_canonical
        Whether to use the mash canonical representation of kmers.

    Returns
    -------
    set[int]
        kmer hashes for the sequence.
    """
    seq = seq.astype(np.int64)

    kmer_hashes = np.zeros(len(seq) - k + 1, dtype=np.int32)

    skip_until = 0
    for i in range(k):
        if seq[i] >= num_states:
            skip_until = i + 1

    kmer_index = 0
    for i in range(len(seq) - k + 1):
        if seq[i + k - 1] >= num_states:
            skip_until = i + k

        if i < skip_until:
            continue
        kmer_hashes[kmer_index] = hash_kmer(seq[i : i + k], mash_canonical)
        kmer_index += 1
    return kmer_hashes[:kmer_index]


@numba.njit
def murmurhash3_32(data: np.ndarray, seed: int = 0x9747B28C) -> int:  # pragma: no cover
    """MurmurHash3 32-bit implementation for an array of integers.

    Parameters
    ----------
    data : np.ndarray
        The input array to hash.
    seed : int
        A seed for the hash function.

    Returns
    -------
    int
        The computed hash value.
    """
    length = data.size
    h = seed ^ length

    for i in range(length):
        k = data[i]

        # Mix the hash
        k *= 0xCC9E2D51
        k = (k << 15) | (k >> (32 - 15))  # Rotate left
        k *= 0x1B873593

        h ^= k
        h = (h << 13) | (h >> (32 - 13))  # Rotate left
        h = h * 5 + 0xE6546B64

    h ^= h >> 16
    h *= 0x85EBCA6B
    h ^= h >> 13
    h *= 0xC2B2AE35
    h ^= h >> 16

    return h & 0xFFFFFFFF  # Return as 32-bit integer


@numba.njit
def hash_kmer(kmer: np.ndarray, mash_canonical: bool) -> int:  # pragma: no cover
    """Hash a kmer, optionally use the mash canonical representaiton.

    Parameters
    ----------
    kmer
        The kmer to hash.
    canonical
        Whether to use the mash canonical representation for a kmer.

    Returns
    -------
    int
        The has of a kmer.

    Notes
    -----
    Uses MurmurHash3 32-bit implementation.
    """
    smallest = kmer
    if mash_canonical:
        reverse = reverse_complement(kmer)
        for i in range(kmer.size):
            if kmer[i] < reverse[i]:
                break
            if kmer[i] > reverse[i]:
                smallest = reverse
                break

    return murmurhash3_32(smallest)


@numba.njit
def reverse_complement(kmer: np.ndarray) -> np.ndarray:  # pragma: no cover
    """Take the reverse complement of a kmer.

    Assumes cogent3 DNA/RNA sequences (numerical
    representation for complement offset by 2
    from original).

    Parameters
    ----------
    kmer
        The kmer to attain the reverse complement of

    Returns
    -------
    numpy.ndarray
        The reverse complement of a kmer.
    """
    # 0123 TCAG
    # 3->1, 1->3, 2->0, 0->2
    return ((kmer + 2) % 4)[::-1]


def mash_distance(
    left_sketch: BottomSketch,
    right_sketch: BottomSketch,
    k: int,
    sketch_size: int,
) -> float:
    """Compute the mash distance between two sketches.

    Parameters
    ----------
    left_sketch
        A sketch for comparison.
    right_sketch
        A sketch for comparison.
    k
        kmer size.
    sketch_size
        Size of the sketches.

    Returns
    -------
    float
        The mash distance between two sketches.
    """
    # Following the source code implementation
    intersection_size = 0
    union_size = 0

    left_index = 0
    right_index = 0
    while (
        union_size < sketch_size
        and left_index < len(left_sketch)
        and right_index < len(right_sketch)
    ):
        left, right = left_sketch[left_index], right_sketch[right_index]
        if left < right:
            left_index += 1
        elif right < left:
            right_index += 1
        else:
            left_index += 1
            right_index += 1
            intersection_size += 1
        union_size += 1

    if union_size < sketch_size:
        if left_index < len(left_sketch):
            union_size += len(left_sketch) - left_index
        if right_index < len(right_sketch):
            union_size += len(right_sketch) - right_index
        union_size = min(union_size, sketch_size)

    jaccard = intersection_size / union_size
    if intersection_size == union_size:
        return 0.0
    if intersection_size == 0:
        return 1.0
    distance = -math.log(2 * jaccard / (1.0 + jaccard)) / k
    if distance > 1:
        distance = 1.0
    return distance


def euclidean_distances(
    seq_arrays: Sequence[SeqArray],
    k: int,
    moltype: str,
    *,
    progress: Progress | None = None,
) -> np.ndarray:
    """Calculates pairwise euclidean distances between sequences.

    Parameters
    ----------
    seqs
        Sequences for pairwise distance calculation.
    progress
        Progress bar, by default None

    Returns
    -------
    np.ndarray
        Pairwise euclidean distances between sequences.
    """
    if progress is None:
        progress = Progress(disable=True)

    kmer_seqs = [
        make_kmerseq(
            seq,
            dtype=np.min_scalar_type(len(_get_canonical_states(moltype)) ** k),
            k=k,
            moltype=moltype,
        )
        for seq in seq_arrays
    ]

    distances = np.zeros((len(kmer_seqs), len(kmer_seqs)))

    distance_task = progress.add_task(
        "[green]Computing Pairwise Distances",
        total=len(kmer_seqs) * (len(kmer_seqs) - 1) // 2,
    )

    for i, kmer_seq_i in enumerate(kmer_seqs):
        freq_i = np.array(kmer_seq_i.kfreqs)
        for j in range(i + 1, len(kmer_seqs)):
            freq_j = np.array(kmer_seqs[j].kfreqs)

            distance = euclidean_distance(freq_i, freq_j)
            distances[i, j] = distance
            distances[j, i] = distance

            progress.update(distance_task, advance=1)

    return distances


def euclidean_distance(freq_1: np.ndarray, freq_2: np.ndarray) -> np.ndarray:
    return np.linalg.norm(freq_1 - freq_2)
